import "Turbine";
import "Turbine.Gameplay";
import "Turbine.UI";
import "Turbine.UI.Lotro";
import "moebius92.ColorBar";
import "moebius92.Handlers";
import "moebius92.Misc";

_G.trim = function(s)
	return string.gsub(s, "^%s*(.-)%s*$", "%1");
end

tempMoraleEffects = { };
tempMoraleEffects["Word of Exaltation"] = true;
tempMoraleEffects["Joy of Battle"] = true;
tempMoraleEffects["Glory of Battle"] = true;
tempMoraleEffects["Sudden Defence"] = true;
tempMoraleEffects["Song of the Hammerhand"] = true;
tempMoraleEffects["Gift of the Hammerhand"] = true;

local baseDurationImageID = 0x41007E35;

effectGroup = { };
effectSlot = { };

categoryString = { };
categoryString[Turbine.Gameplay.EffectCategory.Fear] = "<fear>";
categoryString[Turbine.Gameplay.EffectCategory.Wound] = "<wound>";
categoryString[Turbine.Gameplay.EffectCategory.Disease] = "<disease>";
categoryString[Turbine.Gameplay.EffectCategory.Poison] = "<poison>";

categoryColor = { };
categoryColor[Turbine.Gameplay.EffectCategory.Fear] = Turbine.UI.Color(1.0, 1.0, 0.0, 1.0);
categoryColor[Turbine.Gameplay.EffectCategory.Wound] = Turbine.UI.Color(1.0, 1.0, 0.0, 0.0);
categoryColor[Turbine.Gameplay.EffectCategory.Disease] = Turbine.UI.Color(1.0, 1.0, 1.0, 0.0);
categoryColor[Turbine.Gameplay.EffectCategory.Poison] = Turbine.UI.Color(1.0, 0.0, 1.0, 0.0);

function EffectSlot(list, debuff, icon, name, category)
	local s;
	local r;

	if debuff then
		s = "D";
	else
		s = "B";
	end

	r = list[s .. "/" .. icon .. "/" .. name];
	if r ~= nil then
		return r;
	end

	r = list[s .. "/" .. name];
	if r ~= nil then
		return r;
	end

	r = list[s .. "/" .. icon];
	if r ~= nil then
		return r;
	end

	if categoryString[category] ~= nil then
		r = list[s .. "/" .. icon .. "/" .. categoryString[category]];
		if r ~= nil then
			return r;
		end

		r = list[s .. "/" .. categoryString[category]];
		if r ~= nil then
			return r;
		end
	end

	return nil;
end

classColor = { };
classColor[Turbine.Gameplay.Class.Burglar] =		Turbine.UI.Color(0.3, 0.0, 0.0, 0.0);
classColor[Turbine.Gameplay.Class.Captain] =		Turbine.UI.Color(0.3, 0.0, 0.0, 1.0);
classColor[Turbine.Gameplay.Class.Champion] =		Turbine.UI.Color(0.3, 1.0, 0.0, 0.0);
classColor[Turbine.Gameplay.Class.Guardian] =		Turbine.UI.Color(0.3, 1.0, 0.5, 0.0);
classColor[Turbine.Gameplay.Class.Hunter] =			Turbine.UI.Color(0.3, 0.0, 1.0, 0.0);
classColor[Turbine.Gameplay.Class.LoreMaster] =		Turbine.UI.Color(0.3, 0.0, 1.0, 1.0);
classColor[Turbine.Gameplay.Class.Minstrel] =		Turbine.UI.Color(0.3, 1.0, 1.0, 0.0);
classColor[Turbine.Gameplay.Class.RuneKeeper] =		Turbine.UI.Color(0.3, 1.0, 1.0, 1.0);
classColor[Turbine.Gameplay.Class.Warden] =			Turbine.UI.Color(0.3, 1.0, 0.0, 1.0);

readyColor = { };
readyColor[Turbine.Gameplay.ReadyState.Ready] = Turbine.UI.Color(1.0, 0.0, 1.0, 0.0);
readyColor[Turbine.Gameplay.ReadyState.NotReady] = Turbine.UI.Color(1.0, 1.0, 0.0, 0.0);

PartyMemberWindow = class(Turbine.UI.Control);

function PartyMemberWindow:Constructor(parent, actor, size, width, t)
	Turbine.UI.Control.Constructor(self);

	self.parent = parent;
	self.actor = actor;
	self.t = t;
	if self.t == nil then
		self.class = actor:GetClass();
	else
		self.class = t;
	end
	self.size = size;
	self.width = width;
	self.dead = false;
	self.damage = 0;
	self.rolls = { };
	self.rollOrder = nil;
	self.actorHandle = { };

	self:SetSize(width + size * 11, size);

	self.select = Turbine.UI.Lotro.EntityControl();
	self.select:SetParent(self);
	self.select:SetPosition(0, 0);
	self.select:SetSize(width + size * 11, size);
	self.select:SetMouseVisible(true);
	self.select:SetEntity(actor);
	self.select:SetSelectionEnabled(true);
	self.select:SetContextMenuEnabled(true);

	if classColor[self.class] ~= nil then
		self.color = classColor[self.class];
	else
		self.color = Turbine.UI.Color(0.3, 1.0, 1.0, 1.0)
	end
	self:SetBackColor(self.color);

	self.name = Turbine.UI.Label();
	self.name:SetOutlineColor(Turbine.UI.Color(1.0, 0.0, 0.0, 0.0));
	self.name:SetFontStyle(Turbine.UI.FontStyle.Outline);
	self.name:SetFont(Turbine.UI.Lotro.Font.Verdana16);
	self.name:SetTextAlignment(Turbine.UI.ContentAlignment.MiddleLeft);
	self.name:SetMultiline(false);
	self.name:SetZOrder(100);
	self.name:SetParent(self);
	self.name:SetPosition(size, 0);
	self.name:SetSize(size * 5, size);
	self.name:SetMouseVisible(false);
	self.actorHandle["IsLinkDeadChanged"] = moebius92.Handlers.AddEventHandler(self.actor, "IsLinkDeadChanged", function(sender, args) self:UpdateName(); end);
	self.actorHandle["IsVoiceEnabledChanged"] = moebius92.Handlers.AddEventHandler(self.actor, "IsVoiceEnabledChanged", function(sender, args) self:UpdateName(); end);
	self.actorHandle["NameChanged"] = moebius92.Handlers.AddEventHandler(self.actor, "NameChanged", function(sender, args) self:UpdateName(); end);
	self:UpdateName();

	if self.t == nil then
		self.ready = Turbine.UI.Control();
		self.ready:SetZOrder(100);
		self.ready:SetParent(self);
		self.ready:SetPosition(size / 4, size / 4);
		self.ready:SetSize(size / 2, size / 2);
		self.ready:SetVisible(true);
		self.ready:SetMouseVisible(false);
		self.actorHandle["ReadyStateChanged"] = moebius92.Handlers.AddEventHandler(self.actor, "ReadyStateChanged", function(sender, args) self:UpdateReadyState(); end);
		self.actorHandle["IsVoiceActiveChanged"] = moebius92.Handlers.AddEventHandler(self.actor, "IsVoiceActiveChanged", function(sender, args) self:UpdateReadyState(); end);
		self:UpdateReadyState();

		self.actorHandle["PetChanged"] = moebius92.Handlers.AddEventHandler(self.actor, "PetChanged", function(sender, args) self.parent:PartyMemberChanged(); end);
	else
		self.triangle = Turbine.UI.Control();
		self.triangle:SetZOrder(100);
		self.triangle:SetParent(self);
		self.triangle:SetPosition(0, 0);
		self.triangle:SetSize(size, size);
		self.triangle:SetVisible(true);
		self.triangle:SetMouseVisible(false);
		self.triangle:SetBackground("moebius92/PartyFrame/triangle.tga");
		self.triangle:SetBackColor(Turbine.UI.Color(1.0, self.color.R, self.color.G, self.color.B));
		self.triangle:SetBackColorBlendMode(Turbine.UI.BlendMode.Multiply);
	end

	self.threat = 0;
	self.oldMorale = actor:GetMorale();
	self.moraleBar = moebius92.ColorBar.ColorBar(6, width - 2, (size - 6) / 2);
	self.moraleBar.tempEnd = 0;
	self.moraleBar.tempEffect = { };
	self.moraleBar:SetBarColor(1, Turbine.UI.Color(1, 0.0, 1.0, 0.0));
	self.moraleBar:SetBarColor(2, Turbine.UI.Color(1, 0.5, 0.0, 1.0));
	self.moraleBar:SetBarColor(3, Turbine.UI.Color(1, 1.0, 0.5, 0.0));
	self.moraleBar:SetBarColor(4, Turbine.UI.Color(1, 0.5, 1.0, 0.5));
	self.moraleBar:SetBarColor(5, Turbine.UI.Color(1, 1.0, 0.0, 0.0));
	self.moraleBar:SetParent(self);
	self.moraleBar:SetPosition(size * 6 + 1, 3);
	self.moraleBar:SetMouseVisible(false);
	self.actorHandle["MoraleChanged"] = moebius92.Handlers.AddEventHandler(self.actor, "MoraleChanged", function(sender, args) self:UpdateMorale(); end);
	self.actorHandle["MaxMoraleChanged"] = moebius92.Handlers.AddEventHandler(self.actor, "MaxMoraleChanged", function(sender, args) self:UpdateMorale(); end);
	self.actorHandle["BaseMaxMoraleChanged"] = moebius92.Handlers.AddEventHandler(self.actor, "BaseMaxMoraleChanged", function(sender, args) self:UpdateMorale(); end);
	self.actorHandle["TemporaryMoraleChanged"] = moebius92.Handlers.AddEventHandler(self.actor, "TemporaryMoraleChanged", function(sender, args) self:UpdateMorale(); end);
	self.actorHandle["MaxTemporaryMoraleChanged"] = moebius92.Handlers.AddEventHandler(self.actor, "MaxTemporaryMoraleChanged", function(sender, args) self:UpdateMorale(); end);
	self:UpdateMorale();

	self.powerBar = moebius92.ColorBar.ColorBar(4, width - 2, (size - 6) / 2);
	self.powerBar.tempEnd = 0;
	self.powerBar.tempEffect = { };
	self.powerBar:SetBarColor(1, Turbine.UI.Color(1, 0.0, 0.0, 1.0));
	self.powerBar:SetBarColor(2, Turbine.UI.Color(1, 0.5, 1.0, 0.0));
	self.powerBar:SetBarColor(3, Turbine.UI.Color(1, 1.0, 0.0, 0.5));
	self.powerBar:SetBarColor(4, Turbine.UI.Color(1, 0.5, 0.5, 1.0));
	self.powerBar:SetParent(self);
	self.powerBar:SetPosition(size * 6 + 1, size / 2);
	self.powerBar:SetMouseVisible(false);
	self.actorHandle["PowerChanged"] = moebius92.Handlers.AddEventHandler(self.actor, "PowerChanged", function(sender, args) self:UpdatePower(); end);
	self.actorHandle["MaxPowerChanged"] = moebius92.Handlers.AddEventHandler(self.actor, "MaxPowerChanged", function(sender, args) self:UpdatePower(); end);
	self.actorHandle["BaseMaxPowerChanged"] = moebius92.Handlers.AddEventHandler(self.actor, "BaseMaxPowerChanged", function(sender, args) self:UpdatePower(); end);
	self.actorHandle["TemporaryPowerChanged"] = moebius92.Handlers.AddEventHandler(self.actor, "TemporaryPowerChanged", function(sender, args) self:UpdatePower(); end);
	self.actorHandle["MaxTemporaryPowerChanged"] = moebius92.Handlers.AddEventHandler(self.actor, "MaxTemporaryPowerChanged", function(sender, args) self:UpdatePower(); end);
	self:UpdatePower();

	self.effects = self.actor:GetEffects();
	self.effectsHandle = { };

	self.effectSlots = { };
	for i = -5, 5, 1 do
		if i ~= 0 then
			local x;

			if i < 0 then
				x = (i + 6) * size;
			else
				x = (i + 5) * size + width;
			end
			self.effectSlots[i] = { };
			self.effectSlots[i].icon = Turbine.UI.Control();
			self.effectSlots[i].icon:SetZOrder(101);
			self.effectSlots[i].icon:SetParent(self);
			self.effectSlots[i].icon:SetSize(size, size);
			self.effectSlots[i].icon:SetPosition(x, 0);
			self.effectSlots[i].icon:SetVisible(false);
			self.effectSlots[i].icon:SetBackColor(Turbine.UI.Color(0.7, 1.0, 0.0, 0.0));
			self.effectSlots[i].icon:SetBackColorBlendMode(Turbine.UI.BlendMode.Multiply);
			self.effectSlots[i].display = Turbine.UI.Lotro.EffectDisplay();
			self.effectSlots[i].display:SetZOrder(101);
			self.effectSlots[i].display:SetParent(self);
			self.effectSlots[i].display:SetSize(size, size);
			self.effectSlots[i].display:SetPosition(x, 0);
			self.effectSlots[i].display:SetVisible(false);
			self.effectSlots[i].duration = Turbine.UI.Control();
			self.effectSlots[i].duration:SetPosition(0, 0);
			self.effectSlots[i].duration:SetSize(size, size);
			self.effectSlots[i].duration:SetZOrder(150);
			self.effectSlots[i].duration:SetOpacity(0.7);
			self.effectSlots[i].duration:SetBackColor(Turbine.UI.Color(1.0, 0.0, 0.0, 0.0));
			self.effectSlots[i].duration:SetBackColorBlendMode(Turbine.UI.BlendMode.Multiply);
			self.effectSlots[i].duration:SetMouseVisible(false);
			self.effectSlots[i].duration:SetParent(self.effectSlots[i].display);
			self.effectSlots[i].text = Turbine.UI.Label();
			self.effectSlots[i].text:SetOutlineColor(Turbine.UI.Color(1.0, 0.0, 0.0, 0.0));
			self.effectSlots[i].text:SetFontStyle(Turbine.UI.FontStyle.Outline);
			self.effectSlots[i].text:SetFont(Turbine.UI.Lotro.Font.Verdana12);
			self.effectSlots[i].text:SetTextAlignment(Turbine.UI.ContentAlignment.MiddleCenter);
			self.effectSlots[i].text:SetPosition(0, 0);
			self.effectSlots[i].text:SetSize(size, size);
			self.effectSlots[i].text:SetZOrder(200);
			self.effectSlots[i].text:SetStretchMode(1);
			self.effectSlots[i].text:SetMouseVisible(false);
			self.effectSlots[i].text:SetParent(self.effectSlots[i].display);
			self.effectSlots[i].effects = { };
		end
	end

	self.effectsHandle["EffectAdded"] = moebius92.Handlers.AddEventHandler(self.effects, "EffectAdded", function(sender, args) self:AddEffect(self.effects:Get(args.Index)); end);
	self.effectsHandle["EffectRemoved"] = moebius92.Handlers.AddEventHandler(self.effects, "EffectRemoved", function(sender, args) self:RemoveEffect(args.Effect, true); end);
	self.effectsHandle["EffectsCleared"] = moebius92.Handlers.AddEventHandler(self.effects, "EffectsCleared", function(sender, args) self:ClearEffects(); end);

	self:SetupEffects();

	self:SetParent(parent);
	self:SetWantsUpdates(true);
	self:SetVisible(true);
end

function PartyMemberWindow:SetupEffects()
	self:ClearEffects();
	self.lastEffectUpdate = 0;
	for i = 1, self.effects:GetCount(), 1 do
		self:AddEffect(self.effects:Get(i));
	end
end

function PartyMemberWindow:MouseDown(args)
	self.dragging = true;
	self.safeY = nil;
end

function PartyMemberWindow:MouseMove(args)
	local h = self.size;

	if self.pet ~= nil then
		h = h * 2;
	end

	if self.dragging then
		if self.safeY ~= nil then
			if self.safeY > 0 then
				if args.Y > 0 and args.Y < self.safeY then
					self.safeY = args.Y;
					return;
				else
					self.safeY = nil;
				end
			elseif self.safeY < 0 then
				if args.Y < 0 and args.Y > self.safeY then
					self.safeY = args.Y;
					return;
				else
					self.safeY = nil;
				end
			end
		end

		if args.X < 0 or args.X > (self.width + self.size * 11) then
			local i;

			if args.Y < 0 then
				i = self.indexFellowship - self.parent.settings.orientation;
			elseif args.Y > h then
				i = self.indexFellowship + self.parent.settings.orientation;
			else
				return;
			end

			if i < self.indexFellowship then
				i = i + 1;
			end
			table.remove(self.parent.fellowship[self.indexFellowship], self.indexMember);
			table.insert(self.parent.fellowship, i, { });
			table.insert(self.parent.fellowship[i], self);
			self.dragging = false;
			self.parent:Redraw();
		elseif args.Y < 0 then
			if self.indexMember > 1 then
				local swap = self.parent.fellowship[self.indexFellowship][self.indexMember - 1];

				self.parent.fellowship[self.indexFellowship][self.indexMember] = swap;
				self.parent.fellowship[self.indexFellowship][self.indexMember - 1] = self;
				self.parent:Redraw();
				self.safeY = math.huge;
			elseif args.Y < -(self.size / 2) then
				local i = self.indexFellowship - self.parent.settings.orientation;

				table.remove(self.parent.fellowship[self.indexFellowship], self.indexMember);
				if self.parent.fellowship[i] == nil then
					if i == 0 then
						table.insert(self.parent.fellowship, 1, { });
						i = 1;
					else
						self.parent.fellowship[i] = { };
					end
				end
				table.insert(self.parent.fellowship[i], self);
				self.parent:Redraw();
				self.safeY = math.huge;
			end
		elseif args.Y > h then
			if self.indexMember < #self.parent.fellowship[self.indexFellowship] then
				local swap = self.parent.fellowship[self.indexFellowship][self.indexMember + 1];

				self.parent.fellowship[self.indexFellowship][self.indexMember] = swap;
				self.parent.fellowship[self.indexFellowship][self.indexMember + 1] = self;
				self.parent:Redraw();
				self.safeY = -math.huge;
			elseif args.Y > (h + self.size / 2) then
				local i = self.indexFellowship + self.parent.settings.orientation;

				table.remove(self.parent.fellowship[self.indexFellowship], self.indexMember);
				if self.parent.fellowship[i] == nil then
					if i == 0 then
						table.insert(self.parent.fellowship, 1, { });
						i = 1;
					else
						self.parent.fellowship[i] = { };
					end
				end
				table.insert(self.parent.fellowship[i], 1, self);
				self.parent:Redraw();
				self.safeY = -math.huge;
			end
		end
	end
end

function PartyMemberWindow:MouseUp(args)
	self.dragging = false;
end

function PartyMemberWindow:RemoveHandlers()
	for k, v in pairs(self.actorHandle) do
		moebius92.Handlers.RemoveEventHandler(self.actor, k, v);
	end
	for k, v in pairs(self.effectsHandle) do
		moebius92.Handlers.RemoveEventHandler(self.effects, k, v);
	end
end

function PartyMemberWindow:UpdateName()
	local x = 1.0;
	local name = self.actor:GetName();

	if name == nil or name == "" then
		self.pollName = true;
	else
		self.pollName = false;
	end

	if self.dead then
		x = 0.5;
	end

	if self.rollOrder == nil then
		self.name:SetText(name);
	else
		self.name:SetText("(" .. self.rollOrder .. ") " .. name);
	end

	if self.t ~= nil then
		self.name:SetForeColor(Turbine.UI.Color(1.0, x, x, x));
	elseif self.actor:IsLinkDead() then
		self.name:SetForeColor(Turbine.UI.Color(1.0, x, 0.0, 0.0));
	elseif self.actor:IsVoiceEnabled() then
		self.name:SetForeColor(Turbine.UI.Color(1.0, 0.0, x, 0.0));
	else
		self.name:SetForeColor(Turbine.UI.Color(1.0, x, x, x));
	end
end

function PartyMemberWindow:UpdateReadyState()
	local r = self.actor:GetReadyState();
	local v = self.actor:IsVoiceActive();
	local c = readyColor[r];

	if c == nil then
		if v then
			c = Turbine.UI.Color(1.0, 0.0, 0.0, 1.0);
		else
			c = Turbine.UI.Color(1.0, 1.0, 1.0, 1.0);
		end
	elseif v then
		c = Turbine.UI.Color(1.0, c.R / 2, c.G / 2, 1.0);
	end

	if self.leader then
		self.ready:SetBackColor(c);
	else
		self.ready:SetBackColor(Turbine.UI.Color(1.0, c.R / 2.0, c.G / 2.0, c.B / 2.0));
	end
end

function PartyMemberWindow:UpdateMorale()
	local cur = self.actor:GetMorale();
	local base = self.actor:GetBaseMaxMorale();
	local max = self.actor:GetMaxMorale();
	local temp = self.actor:GetTemporaryMorale();
	local tmax = self.actor:GetMaxTemporaryMorale();
	local damage = self.oldMorale - cur;

	if cur == 0 then
		if not self.dead then
			self.dead = true;
			self:UpdateName();
		end
	else
		if self.dead then
			self.dead = false;
			self:UpdateName();
		end
	end

	if self.parent.inCombat and damage > 0 then
		local time = Turbine.Engine.GetGameTime();

		if self.parent.threatTime > 0 then
			local item = { };

			self.damage = self.damage + damage;
			item.damage = damage;
			item.time = time;
			if self.damageTail ~= nil then
				self.damageTail.next = item;
				self.damageTail = item;
			else
				self.damageHead = item;
				self.damageTail = item;
			end

			while self.damageHead ~= nil do
				if (self.damageHead.time + self.parent.threatTime) >= time then
					break;
				end
				self.damage = self.damage - self.damageHead.damage;
				self.damageHead = self.damageHead.next;
			end

			if self.threat < self.damage then
				self.threat = self.damage;
			end
		end
	end

	local m = math.max(max, base) + temp;
	local min = math.min(max, base);

	self.maxMorale = math.max(max, base) + math.min(tmax, base);
	self.oldMorale = cur;
	
	local t = 1.0;
	if self.moraleBar.tempEnd ~= 0 then
		t = math.min(10.0, math.max(0, self.moraleBar.tempEnd - Turbine.Engine.GetGameTime())) / 10.0;
	end
	self.moraleBar:SetBarColor(2, Turbine.UI.Color(1, 0.5, 0.0, t));

	local mp = math.max(m, self.parent.maxMorale);
	self.moraleBar:SetStats(1, 0, cur, min, mp);
	self.moraleBar:SetStats(2, max, max + temp, max + temp, mp);
	self.moraleBar:SetStats(3, max + temp, m, m, mp);
	self.moraleBar:SetStats(4, base, cur, max, mp);
	self.moraleBar:SetStats(5, 0, cur, math.min(self.threat, max), mp);
	self.parent:UpdateMorale();
end

function PartyMemberWindow:UpdatePower()
	local cur = self.actor:GetPower();
	local base = self.actor:GetBaseMaxPower();
	local max = self.actor:GetMaxPower();
	local temp = self.actor:GetTemporaryPower();
	local tmax = self.actor:GetMaxTemporaryPower();

	local m = math.max(max, base) + temp;

	self.maxPower = math.max(max, base) + math.min(tmax, base);

	local t = 1.0;
	if self.powerBar.tempEnd ~= 0 then
		t = math.min(10.0, math.max(0, self.powerBar.tempEnd - Turbine.Engine.GetGameTime())) / 10.0;
	end
	self.powerBar:SetBarColor(2, Turbine.UI.Color(1, 0.5, t, 0.0));

	local mp = math.max(m, self.parent.maxPower);
	self.powerBar:SetStats(1, 0, cur, max, mp);
	self.powerBar:SetStats(2, max, max + temp, max + temp, mp);
	self.powerBar:SetStats(3, max + temp, m, m, mp);
	self.powerBar:SetStats(4, math.max(0, max - self.parent.powerFeed), cur, max, mp);
	self.parent:UpdatePower();
end

function PartyMemberWindow:UpdateEffectTime(i, time)
	if self.effectSlots[i].durationTime == nil or self.effectSlots[i].durationTime > 1000000 then
		self.effectSlots[i].duration:SetVisible(false);
		return false;
	else
		local x;

		self.effectSlots[i].duration:SetVisible(true);
		if self.effectSlots[i].debuff then
			x = math.floor(math.min(59, math.max(0, 60.0 * (time - (self.effectSlots[i].startTime + self.effectSlots[i].durationTime - 10.0)) / 10.0)));
		else
			x = math.floor(math.min(59, math.max(0, 60.0 * (time - self.effectSlots[i].startTime) / self.effectSlots[i].durationTime)));
		end
		if x ~= self.effectSlots[i].durationIcon then
			self.effectSlots[i].duration:SetBackground(baseDurationImageID + x);
			self.effectSlots[i].duration:SetStretchMode(2);
			self.effectSlots[i].duration:SetStretchMode(1);
			self.effectSlots[i].duration:SetSize(self.size, self.size);
			self.effectSlots[i].durationIcon = x;
		end
		return true;
	end
end

function PartyMemberWindow:UpdateEffect(i)
	local item = self.effectSlots[i].effects[1];

	self.effectSlots[i].durationTime = nil;

	if effectGroup[i] ~= nil then
		local found;

		for _, v in ipairs(effectGroup[i]) do
			found = false;
			for _, v1 in pairs(self.effectSlots[i].effects) do
				if EffectSlot(v.effects, v1.effect:IsDebuff(), v1.effect:GetIcon(), v1.effect:GetName(), v1.effect:GetCategory()) then
					found = true;
					break;
				end
			end

			if not found then
				self.effectSlots[i].icon:SetVisible(true);
				self.effectSlots[i].icon:SetBackground(v.icon);
				self.effectSlots[i].icon:SetStretchMode(2);
				self.effectSlots[i].icon:SetStretchMode(1);
				self.effectSlots[i].icon:SetSize(self.size, self.size);
				break;
			end
		end

		if found then
			self.effectSlots[i].icon:SetVisible(false);
		end

		self.effectSlots[i].display:SetVisible(false);
	elseif item ~= nil then
		local cat = item.effect:GetCategory();
		local c = categoryColor[cat];
		local n = #self.effectSlots[i].effects;
		local name = item.effect:GetName();

		self.effectSlots[i].display:SetEffect(item.effect);
		self.effectSlots[i].startTime = item.effect:GetStartTime();
		self.effectSlots[i].durationTime = item.effect:GetDuration();
		self.effectSlots[i].debuff = item.effect:IsDebuff();

		if not item.effect:IsDebuff() or not item.effect:IsCurable() then
			c = nil;
		end

		if c ~= nil then
			n = 0;
			for _, v in pairs(self.effectSlots[i].effects) do
				if v.effect:IsDebuff() and v.effect:IsCurable() and v.effect:GetCategory() == cat then
					n = n + 1;
				end
			end
			if n > 1 then
				self.effectSlots[i].text:SetText(tostring(n));
				self.effectSlots[i].text:SetForeColor(c);
				self.effectSlots[i].text:SetVisible(true);
			else
				self.effectSlots[i].text:SetVisible(false);
			end
		elseif #name > 6 and string.sub(name, #name - 5, #name - 2) == "Tier" then
			self.effectSlots[i].text:SetText(string.sub(name, #name));
			self.effectSlots[i].text:SetForeColor(Turbine.UI.Color(1.0, 1.0, 1.0, 1.0));
			self.effectSlots[i].text:SetVisible(true);
		elseif n > 1 then
			self.effectSlots[i].text:SetText(tostring(n));
			self.effectSlots[i].text:SetForeColor(Turbine.UI.Color(1.0, 1.0, 1.0, 1.0));
			self.effectSlots[i].text:SetVisible(true);
		else
			self.effectSlots[i].text:SetVisible(false);
		end
		if self:UpdateEffectTime(i, Turbine.Engine.GetGameTime()) then
			self:SetWantsUpdates(true);
		end
		self.effectSlots[i].display:SetVisible(true);
		self.effectSlots[i].icon:SetVisible(false);
	else
		self.effectSlots[i].display:SetEffect(nil);
		self.effectSlots[i].display:SetVisible(false);
		self.effectSlots[i].icon:SetVisible(false);
	end
end

function PartyMemberWindow:AddEffect(effect)
	local id;
	local time = Turbine.Engine.GetGameTime();
	local name = effect:GetName();

	self.parent:LogEffect(effect);

	self.lastEffectUpdate = time;
	for i = 1, self.effects:GetCount(), 1 do
		id = self.effects:Get(i):GetID();

		if self.effectID[id] ~= nil then
			self.effectID[id].time = time;
		end
	end

	id = effect:GetID();

	if tempMoraleEffects[name] then
		self.moraleBar.tempEffect[id] = effect:GetStartTime() + effect:GetDuration();
		self.moraleBar.tempEnd = 0;
		for _, v in pairs(self.moraleBar.tempEffect) do
			self.moraleBar.tempEnd = math.max(self.moraleBar.tempEnd, v);
		end
	end

	if self.effectID[id] == nil then
		local slot = EffectSlot(effectSlot, effect:IsDebuff(), effect:GetIcon(), name, effect:GetCategory());

		if slot ~= nil then
			local item = { };

			item.effect = effect;
			item.time = time;
			item.slot = slot;
			item.startTime = effect:GetStartTime();
			item.endTime = effect:GetStartTime() + effect:GetDuration();
			table.insert(self.effectSlots[slot].effects, item);
			table.sort(self.effectSlots[slot].effects, function(a, b) return a.startTime > b.startTime; end);
			self.effectID[id] = item;
			self:UpdateEffect(slot);
		end
	end
end

function PartyMemberWindow:RemoveEffect(effect, mark)
	local id;
	local time = Turbine.Engine.GetGameTime();
	local name = effect:GetName();

	if mark then
		self.lastEffectUpdate = time;
		for i = 1, self.effects:GetCount(), 1 do
			id = self.effects:Get(i):GetID();

			if self.effectID[id] ~= nil then
				self.effectID[id].time = time;
			end
		end
	end

	id = effect:GetID();

	if tempMoraleEffects[name] then
		self.moraleBar.tempEffect[id] = nil;
		self.moraleBar.tempEnd = 0;
		for _, v in pairs(self.moraleBar.tempEffect) do
			self.moraleBar.tempEnd = math.max(self.moraleBar.tempEnd, v);
		end
	end

	if self.effectID[id] ~= nil then
		local slot = self.effectID[id].slot;

		for k, v in pairs(self.effectSlots[slot].effects) do
			if v == self.effectID[id] then
				table.remove(self.effectSlots[slot].effects, k);
				break;
			end
		end
		self.effectID[id] = nil;
		self:UpdateEffect(slot);
	end
end

function PartyMemberWindow:ClearEffects()
	for i = -5, 5, 1 do
		if i ~= 0 then
			self.effectSlots[i].effects = { };
			self:UpdateEffect(i);
		
		end
	end
	self.effectID = { };

	self.moraleBar.tempEffect = { };
	self.moraleBar.tempEnd = 0;
	self.powerBar.tempEffect = { };
	self.powerBar.tempEnd = 0;
end

function PartyMemberWindow:Update()
	local time = Turbine.Engine.GetGameTime();
	local update = false;

	if self.moraleBar.tempEnd ~= 0 then
		self:UpdateMorale();
	end

	if self.powerBar.tempEnd ~= 0 then
		self:UpdatePower();
	end

	for i = -5, 5, 1 do
		if i ~= 0 and #self.effectSlots[i].effects > 0 then
			if self:UpdateEffectTime(i, time) then
				update = true;
			end
		end
	end

	if self.lastEffectUpdate == time then
		for k, v in pairs(self.effectID) do
			if v.time ~= time then
				self:RemoveEffect(v.effect, false);
			end
		end
	else
		for k, v in pairs(self.effectID) do
			if v.endTime <= time then
				self:RemoveEffect(v.effect, false);
			end
		end
	end

	if self.pollName then
		self:UpdateName();
	end

--	if not update then
--		self:SetWantsUpdates(false);
--	end
end

PartyFrameFTab = class(Turbine.UI.Control);

function PartyFrameFTab:Constructor(parent, index)
	Turbine.UI.Control.Constructor(self);

	self.parent = parent;
	self.index = index;
	self.dragging = false;

	self:SetParent(parent);

	self:SetBackColor(Turbine.UI.Color(1.0, 1.0, 1.0, 1.0));
	self.u = Turbine.UI.Control();
	self.u:SetParent(self);
	self.u:SetPosition(0, 0);
	self.u:SetSize(parent.size, parent.size);
	self.u:SetBackColor(Turbine.UI.Color(1.0, 1.0, 1.0, 1.0));
	self.u:SetBackColorBlendMode(Turbine.UI.BlendMode.Multiply);
	self.u:SetMouseVisible(false);
	self.d = Turbine.UI.Control();
	self.d:SetParent(self);
	self.d:SetSize(parent.size, parent.size);
	self.d:SetBackColor(Turbine.UI.Color(1.0, 1.0, 1.0, 1.0));
	self.d:SetBackColorBlendMode(Turbine.UI.BlendMode.Multiply);
	self.d:SetMouseVisible(false);
end

function PartyFrameFTab:Setup(s, l, orientation)
	if l > 0 then
		self:SetSize(self.parent.size, self.parent.size * l);
		if l == 1 then
			self.d:SetVisible(false);
		else
			self.d:SetPosition(0, self.parent.size * (l - 1));
			self.d:SetVisible(true);
		end
		if orientation == 0 then
			self.u:SetBackground("moebius92/PartyFrame/ur.tga");
			self.d:SetBackground("moebius92/PartyFrame/lr.tga");
			self:SetPosition(self.parent.width + self.parent.size * 11, self.parent.size * s);
		else
			self.u:SetBackground("moebius92/PartyFrame/ul.tga");
			self.d:SetBackground("moebius92/PartyFrame/ll.tga");
			self:SetPosition(0, self.parent.size * s);
		end
		self:SetVisible(true);
	else
		self:SetVisible(false);
	end
end

function PartyFrameFTab:MouseDown(args)
	self.dragging = true;
	self.safeY = nil;
end

function PartyFrameFTab:MouseMove(args)
	if self.dragging then
		if self.safeY ~= nil then
			if self.safeY > 0 then
				if args.Y > 0 and args.Y < self.safeY then
					self.safeY = args.Y;
					return;
				else
					self.safeY = nil;
				end
			elseif self.safeY < 0 then
				if args.Y < 0 and args.Y > self.safeY then
					self.safeY = args.Y;
					return;
				else
					self.safeY = nil;
				end
			end
		end

		if args.Y < -(self.parent.size / 2) then
			local i = self.index - self.parent.settings.orientation;

			if self.parent.fellowtabs[i] ~= nil then
				local x = self.parent.fellowship[self.index];

				self.parent.fellowship[self.index] = self.parent.fellowship[i];
				self.parent.fellowship[i] = x;
				self.parent.fellowtabs[self.index] = self.parent.fellowtabs[i];
				self.parent.fellowtabs[self.index].index = self.index;
				self.parent.fellowtabs[i] = self;
				self.index = i;
				self.parent:Redraw();
				self.safeY = math.huge;
			end
		elseif args.Y > (self:GetHeight() + 0.5 * self.parent.size) then
			local i = self.index + self.parent.settings.orientation;

			if self.parent.fellowtabs[i] ~= nil then
				local x = self.parent.fellowship[self.index];

				self.parent.fellowship[self.index] = self.parent.fellowship[i];
				self.parent.fellowship[i] = x;
				self.parent.fellowtabs[self.index] = self.parent.fellowtabs[i];
				self.parent.fellowtabs[self.index].index = self.index;
				self.parent.fellowtabs[i] = self;
				self.index = i;
				self.parent:Redraw();
				self.safeY = -math.huge;
			end
		end
	end
end

function PartyFrameFTab:MouseUp(args)
	self.dragging = false;
end

PartyFrameTab = class(Turbine.UI.Control);

function PartyFrameTab:Constructor(parent)
	Turbine.UI.Control.Constructor(self);

	self.parent = parent;

	self:SetParent(parent);
	self:SetSize(parent.size * 11 + parent.width, parent.size);
	self:SetBackColor(Turbine.UI.Color(1.0, 1.0, 1.0, 1.0));
	self.l = Turbine.UI.Control();
	self.l:SetParent(self);
	self.l:SetSize(parent.size, parent.size);
	self.l:SetPosition(0, 0);
	self.l:SetBackColor(Turbine.UI.Color(1.0, 1.0, 1.0, 1.0));
	self.l:SetBackColorBlendMode(Turbine.UI.BlendMode.Multiply);
	self.l:SetMouseVisible(false);
	self.r = Turbine.UI.Control();
	self.r:SetParent(self);
	self.r:SetSize(parent.size, parent.size);
	self.r:SetPosition(parent.size * 10 + parent.width, 0);
	self.r:SetBackColor(Turbine.UI.Color(1.0, 1.0, 1.0, 1.0));
	self.r:SetBackColorBlendMode(Turbine.UI.BlendMode.Multiply);
	self.r:SetMouseVisible(false);

	self:ChangedOrientation();
end

function PartyFrameTab:ChangedOrientation()
	if self.parent.settings.orientation == -1 then
		self.l:SetBackground("moebius92/PartyFrame/ll.tga");
		self.r:SetBackground("moebius92/PartyFrame/lr.tga");
	else
		self.l:SetBackground("moebius92/PartyFrame/ul.tga");
		self.r:SetBackground("moebius92/PartyFrame/ur.tga");
	end
end

function PartyFrameTab:MouseDown(args)
	if args.Button == Turbine.UI.MouseButton.Left then
		self.dragStartX = args.X;
		self.dragStartY = args.Y;
		self.parent.dragging = true;
	end
end

function PartyFrameTab:MouseMove(args)
	if self.parent.dragging then
		self.parent.settings.x = self.parent.settings.x + (args.X - self.dragStartX);
		self.parent.settings.y = self.parent.settings.y + (args.Y - self.dragStartY);

		if self.parent.settings.x < 0 then
			self.parent.settings.x = 0;
		elseif self.parent.settings.x > (self.parent.w - (self.parent.size * 11 + self.parent.width)) then
			self.parent.settings.x = self.parent.w - (self.parent.size * 11 + self.parent.width);
		end

		self.parent:Redraw();
	end
end

function PartyFrameTab:MouseUp(args)
	if args.Button == Turbine.UI.MouseButton.Left and self.parent.dragging then
		self.parent.dragging = false;
		self.parent:SaveSettings();
	end
end

function PartyFrameTab:MouseClick(args)
	if args.Button == Turbine.UI.MouseButton.Right then
		self.parent.settings.orientation = -self.parent.settings.orientation;
		self:ChangedOrientation();
		self.parent:Redraw();
		self.parent:SaveSettings();
	end
end

PartyFrameWindow = class(Turbine.UI.Window);

function PartyFrameWindow:Constructor(size, width)
	Turbine.UI.Window.Constructor(self);

	self.move = false;
	self.dragging = false;
	self.size = size;
	self.width = width;
	self.w = Turbine.UI.Display:GetWidth();
	self.h = Turbine.UI.Display:GetHeight();

	self:LoadSettings();

	self.visible = true;
	self.hide = false;
	self:SetWantsKeyEvents(true);

	self.threatTime = 5;
	self.powerFeed = 1400;
	self.player = Turbine.Gameplay.LocalPlayer.GetInstance();
	self.class = self.player:GetClass();
	self.effects = self.player:GetEffects();
	self.maxMorale = 1;
	self.maxPower = 1;
	self.inCombat = self.player:IsInCombat();

	self.players = { };
	self.actors = { };
	self.partyHandle = { };
	self.fellowship = { };

	self.tab = PartyFrameTab(self);
	self.fellowtabs = { };

	self:UpdateParty();
	moebius92.Handlers.AddEventHandler(self.player, "InCombatChanged", function(sender, args) self:CombatChange(); end);
	moebius92.Handlers.AddEventHandler(self.player, "PartyChanged", function(sender, args) self:UpdateParty(); end);
	moebius92.Handlers.AddChatHandler(function(m) self:ChatMessage(m) end, Turbine.ChatType.Standard, nil);

	self:PetChanged();
	moebius92.Handlers.AddEventHandler(self.player, "PetChanged", function(sender, args) self:PetChanged(); end);
	moebius92.Handlers.AddEventHandler(self.effects, "EffectAdded", function(sender, args) self:LogEffect(self.effects:Get(args.Index)); end);
end

function PartyFrameWindow:LogEffect(effect)
	local debuff = effect:IsDebuff();
	local icon = effect:GetIcon();
	local name = effect:GetName();

	if self.options == nil then
		return;
	end

	if (debuff and self.options.logDebuffs.index == 2) or (not debuff and self.options.logBuffs.index == 2) then
		if self.options.logType.index == 3 or (EffectSlot(effectSlot, debuff, icon, name, effect:GetCategory()) == nil and self.options.logType.index == 2) then
			for _, v in pairs(self.options.slots[0]) do
				if v.name == name and v.debuff == debuff and v.icon == icon then
					return;
				end
			end

			table.insert(self.options.slots[0], { name = name, debuff = debuff, icon = icon });
			table.sort(self.options.slots[0], function(a, b) return a.name < b.name; end);
			self.options:SizeChanged();
		end
	end
end

function PartyFrameWindow:PetChanged()
	if self.pet ~= nil then
		moebius92.Handlers.RemoveEventHandler(self.petEffects, "EffectAdded", self.petHandler);
		self.pet = nil;
	end

	self.pet = self.player:GetPet();

	if self.pet ~= nil then
		self.petEffects = self.pet:GetEffects();
		self.petHandler = moebius92.Handlers.AddEventHandler(self.petEffects, "EffectAdded", function(sender, args) self:LogEffect(self.petEffects:Get(args.Index)); end);
	end
end

function PartyFrameWindow:SetVisible(v)
	if v ~= nil then
		self.visible = v;
	end
	Turbine.UI.Window.SetVisible(self, self.visible and not self.hide);
end

function PartyFrameWindow:LoadSettings()
	self.settings = Turbine.PluginData.Load(Turbine.DataScope.Character, "PartyFrame");

	if self.settings == nil then
		self.settings = { };
		self.settings.x = self.w - (self.size * 11 + self.width);
		self.settings.y = self.h / 2;
		self.settings.orientation = -1;
	end
end

function PartyFrameWindow:SaveSettings()
	Turbine.PluginData.Save(Turbine.DataScope.Character, "PartyFrame", self.settings);
end

function PartyFrameWindow:KeyDown(args)
	if args.Action == 0x1000007B then
		if self.move then
			if self.dragging then
				self.dragging = false;
				self:SaveSettings();
			end
			for _, v in pairs(self.fellowtabs) do
				v.dragging = false;
			end
			for k, v in pairs(self.players) do
				v.dragging = false;
				v.select:SetMouseVisible(true);
			end
			self.move = false;
		else
			for _, v in pairs(self.players) do
				v.select:SetMouseVisible(false);
			end
			self.move = true;
		end
		self:Redraw();
	elseif args.Action == 0x100000B3 then
		self.hide = not self.hide;
		self:SetVisible();
	end
end

function PartyFrameWindow:ChatMessage(m)
	local i = string.find(m, " ");

	if i ~= nil then
		local name = string.sub(m, 1, i - 1);

		m = string.sub(m, i + 1, -1);
		if string.sub(m, 1, 14) == "rolled (100): " then
			if self.players[name] then
				local t = { };
				local i, j, k;
				local state, w;

				table.insert(self.players[name].rolls, tonumber(string.sub(m, 15, -1)));
				self.players[name]:UpdateName();

				for k, v in pairs(self.players) do
					table.insert(t, { ["rolls"] = v.rolls, ["name"] = k });
				end
				table.sort(t, 
function(a, b)
	local n = math.min(#a.rolls, #b.rolls);
	local i;

	for i = 1, math.min(#a.rolls, #b.rolls), 1 do
		if a.rolls[i] > b.rolls[i] then
			return true;
		elseif a.rolls[i] < b.rolls[i] then
			return false;
		end
	end

	return #a.rolls > #b.rolls;
end
				          );

				k = 0;
				for i = 1, #t, 1 do
					if #t[i].rolls == 0 then
						break;
					end

					state = 0;
					if i == 1 then
						state = 2
					else
						for j = 1, #t[i].rolls, 1 do
							if t[i - 1].rolls[j] > t[i].rolls[j] then
								state = 2;
								break;
							end
						end

						if state == 0 then
							if #t[i - 1].rolls > #t[i].rolls then
								state = 1;
							end
						end
					end

					if state == 2 then
						k = k + 1;
					end
					
					if state == 1 then
						w = k .. "?";
					else
						w = k
					end

					if self.players[t[i].name].rollOrder ~= w then
						self.players[t[i].name].rollOrder = w;
						self.players[t[i].name]:UpdateName();
					end
				end
			end
		end
	end
end

function PartyFrameWindow:AddMember(member, name)
	if self.players[name] == nil then
		self.players[name] = PartyMemberWindow(self, member, self.size, self.width, nil);
	end

	if not self.players[name].exists then
		self.players[name].exists = true;
		self.players[name]:SetVisible(true);
		self.players[name]:UpdateReadyState();
		table.insert(self.actors, self.players[name]);

		local pet = member:GetPet();

		if self.players[name].pet == nil or self.players[name].pet.actor ~= pet then
			if self.players[name].pet ~= nil then
				self.players[name].pet:RemoveHandlers();
				self.players[name].pet:SetVisible(false);
				self.players[name].pet:SetParent(nil);
				self.players[name].pet = nil;
			end

			if pet ~= nil then
				self.players[name].pet = PartyMemberWindow(self, pet, self.size, self.width, self.players[name].class);
			end
		end

		if self.players[name].pet ~= nil then
			self.players[name].pet:SetVisible(true);
			table.insert(self.actors, self.players[name].pet);
		end
	end
end

function PartyFrameWindow:UpdateParty()
	for k, v in pairs(self.partyHandle) do
		moebius92.Handlers.RemoveEventHandler(self.party, k, v);
	end
	self.partyHandle = { };

	for k, v in pairs(self.actors) do
		v:RemoveHandlers();
		v:SetVisible(false);
		v:SetParent(nil);
	end
	self.players = { };
	self.actors = { };

	self.party = self.player:GetParty();
	if self.party ~= nil then
		self.partyHandle["MemberAdded"] = moebius92.Handlers.AddEventHandler(self.party, "MemberAdded", function(sender, args) self:PartyMemberChanged(); end);
		self.partyHandle["MemberRemoved"] = moebius92.Handlers.AddEventHandler(self.party, "MemberRemoved", function(sender, args) self:PartyMemberChanged(); end);
		self.partyHandle["LeaderChanged"] = moebius92.Handlers.AddEventHandler(self.party, "LeaderChanged", function(sender, args) self:PartyMemberChanged(); end);
	end

	self:PartyMemberChanged();
end

function PartyFrameWindow:PartyMemberChanged(order)
	self.fellowship = { };
	if self.party ~= nil and self.party:GetMemberCount() >= 1 then
		local n = self.party:GetMemberCount();
		local l = self.party:GetLeader();
		local lname = l:GetName();
		local members = { };
		local name = self.player:GetName();

		self.actors = { };

		for k, v in pairs(self.players) do
			v.exists = false;
			v.leader = false;
		end

		for i = 1, n, 1 do
			local member = self.party:GetMember(i);

			members[member:GetName()] = { index = i, member = member };
		end
		if members[name] == nil then
			members[name] = { index = 0, member = self.player };
		end

		if order ~= nil then
			local f = 1;
			local m = 1;

			for v in string.gmatch(order, "[^ ]+") do
				v = trim(v);
				self.fellowship[f] = { };
				m = 1;
				for v1 in string.gmatch(v, "[^,]+") do
					if members[v1] ~= nil then
						self.fellowship[f][m] = members[v1].member;
						m = m + 1;
						members[v1] = nil;
					else
						Turbine.Shell.WriteLine(v1 .. " not found.");
					end
				end

				if m > 1 then
					f = f + 1;
				end
			end

			local t = { };
			for k, v in pairs(members) do
				table.insert(t, { index = v.index, member = v.member });
			end
			table.sort(t, function(a, b) return a.index < b.index; end);

			self.fellowship[f] = { };
			m = 1;
			for _, v in ipairs(t) do
				self.fellowship[f][m] = v.member;
				m = m + 1;
			end
		else
			local t;
			local f = 1;
			local m = 1;

			self.fellowship[f] = { };

			if (self.players[name] == nil or self.players[name].indexFellowship == nil) and members[name] ~= nil then
				self.fellowship[f][m] = members[name].member;
				members[name] = nil;
				m = m + 1;
			end

			if (self.players[lname] == nil or self.players[lname].indexFellowship == nil) and members[lname] ~= nil then
				self.fellowship[f][m] = members[lname].member;
				members[lname] = nil;
				m = m + 1;
			end

			t = { };
			for k, v in pairs(members) do
				if self.players[k] ~= nil and self.players[k].indexFellowship ~= nil then
					table.insert(t, { member = v.member, f = self.players[k].indexFellowship, m = self.players[k].indexMember });
					members[k] = nil;
				end
			end

			if #t > 0 then
				local last;

				table.sort(t, function(a, b) if a.f == b.f then return a.m < b.m; else return a.f < b.f; end end);
				last = t[1].f;
				for k, v in ipairs(t) do
					if v.f ~= last then
						f = f + 1;
						m = 1;
						last = v.f;
						self.fellowship[f] = { };
					end
					self.fellowship[f][m] = v.member;
					m = m + 1;
				end
			end

			t = { };
			for k, v in pairs(members) do
				table.insert(t, { index = v.index, member = v.member });
			end
			table.sort(t, function(a, b) return a.index < b.index; end);

			for _, v in ipairs(t) do
				self.fellowship[f][m] = v.member;
				m = m + 1;
			end
		end

		for i = 1, #self.fellowship, 1 do
			for k, v in pairs(self.fellowship[i]) do
				name = v:GetName();
				self:AddMember(v, name);
				self.fellowship[i][k] = self.players[name];
			end
		end

		if self.players[lname] ~= nil then
			self.players[lname].leader = true;
			self.players[lname]:UpdateReadyState();
		end

		for k, v in pairs(self.players) do
			if not v.exists then
				v:RemoveHandlers();
				v:SetVisible(false);
				v:SetParent(nil);
				if v.pet ~= nil then
					v.pet:RemoveHandlers();
					v.pet:SetVisible(false);
					v.pet:SetParent(nil);
				end
				self.players[k] = nil;
			end
		end

		self:UpdateMorale();
		self:UpdatePower();
		for _, v in pairs(self.actors) do
			v:UpdateMorale();
			v:UpdatePower();
		end
	else
		for k, v in pairs(self.players) do
			v:RemoveHandlers();
			v:SetVisible(false);
			v:SetParent(nil);
			if v.pet ~= nil then
				v.pet:RemoveHandlers();
				v.pet:SetVisible(false);
				v.pet:SetParent(nil);
			end
			self.players[k] = nil;
		end
	end
	self:Redraw();
end

function PartyFrameWindow:Redraw()
	local j = 0;
	local x = 0;
	local s = 0;
	local l = 0;
	local t = { };

	for i = 1, #self.fellowship, 1 do
		if #self.fellowship[i] == 0 then
			table.insert(t, i);
		end
	end
	table.sort(t, function(a, b) return a > b; end);
	for _, v in ipairs(t) do
		table.remove(self.fellowship, v);
	end

	if self.move and self.settings.x >= (self.w / 2) then
		x = self.size;
	end

	for _, v in pairs(self.fellowtabs) do
		v:Setup(0, 0, x);
	end

	if self.settings.orientation == -1 then
		for i = #self.fellowship, 1, -1 do
			s = j;
			for k, v in ipairs(self.fellowship[i]) do
				v:SetPosition(x, j * self.size);
				v.indexFellowship = i;
				v.indexMember = k;
				j = j + 1;
				if v.pet ~= nil then
					v.pet:SetPosition(x, j * self.size);
					j = j + 1;
				end
			end

			if self.move then
				if self.fellowtabs[i] == nil then
					self.fellowtabs[i] = PartyFrameFTab(self, i);
				end
				l = j - s;
				self.fellowtabs[i]:Setup(s, l, x);
			end

			j = j + 1;
		end
		if #self.fellowship > 0 then
			j = j - 1;
		end

		self:SetPosition(self.settings.x - x, math.max(0, self.settings.y - self.size * j));
		if self.move then
			self.tab:SetPosition(x, j * self.size);
			j = j + 1;
			self.tab:SetVisible(true);
		else
			self.tab:SetVisible(false);
		end
	else
		if self.move then
			j = j + 1;
		end
		for i = 1, #self.fellowship, 1 do
			s = j;
			for k, v in ipairs(self.fellowship[i]) do
				v:SetPosition(x, j * self.size);
				v.indexFellowship = i;
				v.indexMember = k;
				j = j + 1;
				pet = (v.pet ~= nil);
				if pet then
					v.pet:SetPosition(x, j * self.size);
					j = j + 1;
				end
			end

			if self.move then
				if self.fellowtabs[i] == nil then
					self.fellowtabs[i] = PartyFrameFTab(self, i);
				end
				l = j - s;
				if pet and x ~= 0 then
					l = l - 1;
				end
				self.fellowtabs[i]:Setup(s, l, x);
			end

			j = j + 1;
		end
		if #self.fellowship > 0 then
			j = j - 1;
		end

		if self.move then
			self:SetPosition(self.settings.x - x, math.min(self.h - self.size * j, self.settings.y - self.size));
			self.tab:SetPosition(x, 0);
			self.tab:SetVisible(true);
		else
			self:SetPosition(self.settings.x - x, math.min(self.h - self.size * j, self.settings.y));
			self.tab:SetVisible(false);
		end
	end

	if j > 0 then
		if self.move then
			self:SetSize(self.size * 12 + self.width, self.size * j);
		else
			self:SetSize(self.size * 11 + self.width, self.size * j);
		end
		self:SetVisible(true);
	else
		self:SetVisible(false);
	end
end

function PartyFrameWindow:UpdateMorale()
	local m = 1;

	for k, v in pairs(self.actors) do
		m = math.max(m, v.maxMorale);
	end

	if m ~= self.maxMorale then
		self.maxMorale = m;
		for k, v in pairs(self.actors) do
			v:UpdateMorale();
		end
	end
end

function PartyFrameWindow:UpdatePower()
	local m = 1;

	for k, v in pairs(self.actors) do
		m = math.max(m, v.maxPower);
	end

	if m ~= self.maxPower then
		self.maxPower = m;
		for k, v in pairs(self.actors) do
			v:UpdatePower();
		end
	end
end

function PartyFrameWindow:CombatChange()
	self.inCombat = self.player:IsInCombat();
	for k, v in pairs(self.actors) do
		v.damageHead = nil;
		v.damageTail = nil;
		v.damage = 0;
		v.threat = 0;
		v:UpdateMorale();
	end
end

PartyFrameOptionsButton = class(Turbine.UI.Control);

function PartyFrameOptionsButton:Constructor(parent, index, missing)
	Turbine.UI.Control.Constructor(self);

	self.parent = parent;
	self.index = index;
	self.highlight = false;
	self.missing = missing;

	self:SetParent(parent);
	self:SetSize(parent.window.size - 2, parent.window.size - 2);
	self:Colorize();
end

function PartyFrameOptionsButton:SetMissing(m)
	if m ~= self.missing then
		if m then
			local t = self.parent.slots[self.index];

			self.parent.slots[self.index] = { };
			if #t > 0 then
				self.parent.slots[self.index][1] = t;
			end
		else
			local t = self.parent.slots[self.index];

			self.parent.slots[self.index] = { };
			for _, v in pairs(t) do
				for _, v1 in pairs(v) do
					table.insert(self.parent.slots[self.index], v1);
				end
			end
			table.sort(self.parent.slots[self.index], function(a, b) return a.name < b.name; end);
		end

		self.missing = m;
		self:Colorize();
		self.parent:Changed();
	end
end

function PartyFrameOptionsButton:Colorize()
	local c = 0.5;

	if self.parent.select == self.index then
		c = 1.0;
	end

	if self.missing then
		self:SetBackColor(Turbine.UI.Color(1.0, c, 0.0, 0.0));
	else
		self:SetBackColor(Turbine.UI.Color(1.0, c, c, c));
	end
end

function PartyFrameOptionsButton:MouseClick(args)
	self.parent:Select(self.index);
end

function PartyFrameOptionsButton:MouseEnter(args)
	if self.parent.dragSource ~= nil and self.parent.dragTime == Turbine.Engine.GetGameTime() then
		local x, y = self.parent:GetIndices(self.parent.dragSource.index);
		local t;

		if y == nil then
			t = self.parent.slots[self.parent.select][x];
			if self.parent.select ~= 0 then
				table.remove(self.parent.slots[self.parent.select], x);
			end
		else
			t = self.parent.slots[self.parent.select][x][y];
			if self.parent.select ~= 0 then
				table.remove(self.parent.slots[self.parent.select][x], y);
				if #self.parent.slots[self.parent.select][x] == 0 then
					table.remove(self.parent.slots[self.parent.select], x);
				end
			end
		end

		if self.missing then
			table.insert(self.parent.slots[self.index], { [1] = t });
		else
			table.insert(self.parent.slots[self.index], t);
			table.sort(self.parent.slots[self.index], function(a, b) return (not a.name and "" or a.name) < (not b.name and "" or b.name); end);
		end

		self.parent:Changed();
	end
end

PartyFrameOptionsDropDown = class(Turbine.UI.Lotro.Button)

function PartyFrameOptionsDropDown:Constructor(parent, index, func, ...)
	Turbine.UI.Lotro.Button.Constructor(self);

	self.parent = parent;
	self.label = arg;
	self.func = func;
	self.items = { };

	self:SetParent(parent);
	self.dropdown = Turbine.UI.ContextMenu();

	local items = self.dropdown:GetItems();
	for k, v in ipairs(self.label) do
		self.items[k] = PartyFrameOptionsDDItem(v, self, k);
		items:Add(self.items[k]);
	end

	self:SetIndex(index);
end

function PartyFrameOptionsDropDown:SetIndex(index, click)
	for k, v in ipairs(self.items) do
		v:SetChecked(k == index);
	end
	self:SetText(self.label[index]);
	self.index = index;
	if click and self.func ~= nil then
		self.func(self, self.index);
	end
end

function PartyFrameOptionsDropDown:MouseClick(args)
	self.dropdown:ShowMenuAt(self:PointToScreen(0, self:GetHeight()));
end

PartyFrameOptionsDDItem = class(Turbine.UI.MenuItem)

function PartyFrameOptionsDDItem:Constructor(label, parent, index)
	Turbine.UI.MenuItem.Constructor(self, label);

	self.parent = parent;
	self.index = index;
end

function PartyFrameOptionsDDItem:Click(args)
	self.parent:SetIndex(self.index, true);
end

PartyFrameOptionsEffectText = class(Turbine.UI.Lotro.TextBox);

function PartyFrameOptionsEffectText:Constructor(parent)
	Turbine.UI.Lotro.TextBox.Constructor(self);

	self.parent = parent;
	self:SetParent(parent);
	self:SetMultiline(false);
	self:SetTextAlignment(Turbine.UI.ContentAlignment.MiddleLeft);
	self:SetFont(Turbine.UI.Lotro.Font.Verdana16);
	self:SetWantsKeyEvents(true);
end

function PartyFrameOptionsEffectText:KeyDown(args)
	if self:HasFocus() and args.Action == 162 then
		local name = self:GetText();
		local option = self.parent.parent;
		local x, y = option:GetIndices(self.parent.index);

		self.parent:Focus();
		if option.slotsButton[option.select].missing then
			if x == nil then
				table.insert(option.slots[option.select], { [1] = { name = name, debuff = false } });
			else
				option.slots[option.select][x][y].name = name;
			end
		else
			if x == nil then
				table.insert(option.slots[option.select], { name = name, debuff = false });
			else
				option.slots[option.select][x].name = name;
			end
			table.sort(option.slots[option.select], function(a, b) return (not a.name and "" or a.name) < (not b.name and "" or b.name); end);
		end
		option:Changed();
	end
end

PartyFrameOptionsEffect = class(Turbine.UI.Control)

function PartyFrameOptionsEffect:Constructor(parent, index, icon, debuff, name, ro)
	Turbine.UI.Control.Constructor(self);

	self.parent = parent;
	self.index = index;
	self.dragging = false;
	self:SetParent(parent);
	self:SetMouseVisible(true);

	self.icon = Turbine.UI.Control();
	self.icon:SetParent(self);
	self.icon:SetMouseVisible(false);
	self.icon:SetBackColor(Turbine.UI.Color(1.0, 0.0, 0.0, 0.0));
	self.icon:SetPosition(1, 1);
	self.textbox = PartyFrameOptionsEffectText(self);
	self:SetIcon(icon);
	self:SetDebuff(debuff);
	self:SetName(name, ro);
end

function PartyFrameOptionsEffect:MouseDown(args)
	if args.Button == Turbine.UI.MouseButton.Left then
		local x, y = Turbine.UI.Display.GetMousePosition();

		self.parent.dragWindow:SetBackground(self.image);
		if self.image ~= nil then
			self.parent.dragWindow:SetBackColorBlendMode(Turbine.UI.BlendMode.Screen);
		else
			self.parent.dragWindow:SetBackColorBlendMode(Turbine.UI.BlendMode.Normal);
		end
		self.parent.dragWindow:Activate();
		self.parent.dragWindow:SetZOrder(5000);
		self.parent.dragWindow:SetVisible(true);
		self.parent.dragWindow:SetPosition(x - 16, y - 16);
		self.dragging = true;
	end
end

function PartyFrameOptionsEffect:MouseMove(args)
	if self.dragging then
		local x, y = Turbine.UI.Display.GetMousePosition();
		self.parent.dragWindow:SetPosition(x - 16, y - 16);
	end
end

function PartyFrameOptionsEffect:MouseEnter(args)
	if self.parent.dragSource ~= nil and self.parent.dragTime == Turbine.Engine.GetGameTime() and self.parent.select ~= 0 then
		local tx, ty = self.parent:GetIndices(self.index);
		local sx, sy = self.parent:GetIndices(self.parent.dragSource.index);
		local x, y = self:GetMousePosition();
		local d = (y > (self.parent.window.size / 2)) and 1 or 0;

		if sx == nil then
			if tx == nil then
				if self.parent.slotsButton[self.parent.select].missing then
					table.insert(self.parent.slots[self.parent.select], { [1] = { debuff = false } });
				else
					table.insert(self.parent.slots[self.parent.select], { debuff = false });
				end
			elseif ty == nil then
				table.insert(self.parent.slots[self.parent.select], tx + d, { debuff = false });
			else
				table.insert(self.parent.slots[self.parent.select][tx], ty + d, { debuff = false });
			end
			self.parent:Changed();
		elseif self ~= self.parent.dragSource and tx ~= nil and self.parent.slotsButton[self.parent.select].missing then
			if sx ~= tx or sy < ty then
				table.insert(self.parent.slots[self.parent.select][tx], ty + d, self.parent.slots[self.parent.select][sx][sy]);
				table.remove(self.parent.slots[self.parent.select][sx], sy);
				if #self.parent.slots[self.parent.select][sx] == 0 then
					table.remove(self.parent.slots[self.parent.select], sx);
				end
			else
				local t = self.parent.slots[self.parent.select][sx][sy];

				table.remove(self.parent.slots[self.parent.select][sx], sy);
				table.insert(self.parent.slots[self.parent.select][tx], ty + d, t);
				if #self.parent.slots[self.parent.select][sx] == 0 then
					table.remove(self.parent.slots[self.parent.select], sx);
				end
			end
			self.parent:Changed();
		end
	end
end

function PartyFrameOptionsEffect:MouseUp(args)
	if args.Button == Turbine.UI.MouseButton.Left then
		self.parent.dragWindow:SetVisible(false);
		self.dragging = false;
		self.parent.dragSource = self;
		self.parent.dragTime = Turbine.Engine.GetGameTime();
	end
end

function PartyFrameOptionsEffect:MouseClick(args)
	if args.Button == Turbine.UI.MouseButton.Right then
		local x, y = self.parent:GetIndices(self.index);

		self.parent.effectsMenuTarget = self;
		self.parent.effectsOptions.Debuff:SetChecked(self.debuff);
		self.parent.effectsIndex = { x = x, y = y };
		if self.parent.select == 0 then
			self.parent.effectsOptions.Cut:SetEnabled(false);
			self.parent.effectsOptions.Copy:SetEnabled(true);
			self.parent.effectsOptions.Paste:SetEnabled(false);
			self.parent.effectsOptions.Insert:SetEnabled(false);
			self.parent.effectsOptions.Delete:SetEnabled(false);
			self.parent.effectsOptions.Debuff:SetEnabled(false);
			self.parent.effectsOptions.Save:SetEnabled(true);
			self.parent.effectsOptions.NewGroup:SetEnabled(false);
		else
			self.parent.effectsOptions.Cut:SetEnabled(x ~= nil);
			self.parent.effectsOptions.Copy:SetEnabled(x ~= nil);
			self.parent.effectsOptions.Paste:SetEnabled(self.parent.effectsClipboard ~= nil);
			self.parent.effectsOptions.Insert:SetEnabled(true);
			self.parent.effectsOptions.Delete:SetEnabled(x ~= nil);
			self.parent.effectsOptions.Debuff:SetEnabled(x ~= nil);
			self.parent.effectsOptions.Save:SetEnabled(x ~= nil);
			self.parent.effectsOptions.NewGroup:SetEnabled(self.parent.slotsButton[self.parent.select].missing and x ~= nil and y > 1);
		end
		self.parent.effectsMenu:ShowMenu();
	end
end

function PartyFrameOptionsEffect:SetIcon(icon)
	local w, h = self:GetSize();

	self.icon:SetBackground(icon);
	self.image = icon;
	if icon == nil then
		self.icon:SetBackColorBlendMode(Turbine.UI.BlendMode.Normal);
	else
		self.icon:SetBackColorBlendMode(Turbine.UI.BlendMode.Screen);
	end
end

function PartyFrameOptionsEffect:SetName(name, ro)
	self.textbox:SetText(name);
	self.textbox:SetReadOnly(ro);
end

function PartyFrameOptionsEffect:SetDebuff(debuff)
	self.debuff = debuff;
	if debuff then
		self.textbox:SetForeColor(Turbine.UI.Color(1.0, 1.0, 0.0, 0.0));
	else
		self.textbox:SetForeColor(Turbine.UI.Color(1.0, 0.0, 1.0, 0.0));
	end
end

function PartyFrameOptionsEffect:SizeChanged(args)
	local w, h = self:GetSize();

	self.icon:SetSize(h - 2, h - 2);
	self.textbox:SetSize(w - (h + 2), h - 2);
	self.textbox:SetPosition(h + 1, 1);
end

PartyFrameOptionsGroup = class(Turbine.UI.Control);

function PartyFrameOptionsGroup:Constructor(parent, index)
	Turbine.UI.Control.Constructor(self);

	self.parent = parent;
	self.index = index;
	self.dragging = false;

	self:SetParent(parent);

	self.u = Turbine.UI.Control();
	self.u:SetParent(self);
	self.u:SetPosition(0, 0);
	self.u:SetSize(parent.window.size, parent.window.size);
	self.u:SetBackground("moebius92/PartyFrame/ul.tga");
	self.u:SetBlendMode(Turbine.UI.BlendMode.Overlay);
	self.u:SetMouseVisible(false);
	self.c = Turbine.UI.Control();
	self.c:SetParent(self);
	self.c:SetSize(parent.window.size, parent.window.size);
	self.c:SetBackColor(Turbine.UI.Color(1.0, 1.0, 1.0, 1.0));
	self.c:SetPosition(0, parent.window.size);
	self.c:SetMouseVisible(false);
	self.d = Turbine.UI.Control();
	self.d:SetParent(self);
	self.d:SetSize(parent.window.size, parent.window.size);
	self.d:SetBackground("moebius92/PartyFrame/ll.tga");
	self.d:SetBlendMode(Turbine.UI.BlendMode.Overlay);
	self.d:SetMouseVisible(false);
end

function PartyFrameOptionsGroup:Setup(l)
	if l > 0 then
		self:SetSize(self.parent.window.size, (self.parent.window.size + 2) * l - 2);

		if l == 1 then
			self.c:SetVisible(false);
			self.d:SetVisible(false);
		else
			self.d:SetPosition(0, (self.parent.window.size + 2) * (l - 1) - 2);
			self.d:SetVisible(true);
			self.c:SetSize(self.parent.window.size, (self.parent.window.size + 2) * (l - 2));
			self.c:SetVisible(true);
		end

		self:SetVisible(true);
	else
		self:SetVisible(false);
	end
end

function PartyFrameOptionsGroup:MouseDown(args)
	self.dragging = true;
	self.safeY = nil;
end

function PartyFrameOptionsGroup:MouseMove(args)
	if self.dragging then
		if self.safeY ~= nil then
			if self.safeY > 0 then
				if args.Y > 0 and args.Y < self.safeY then
					self.safeY = args.Y;
					return;
				else
					self.safeY = nil;
				end
			elseif self.safeY < 0 then
				if args.Y < 0 and args.Y > self.safeY then
					self.safeY = args.Y;
					return;
				else
					self.safeY = nil;
				end
			end
		end

		if args.Y < 0 then
			local i = self.index - 1;

			if self.parent.groups[i] ~= nil then
				local x = self.parent.slots[self.parent.select][self.index];

				self.parent.slots[self.parent.select][self.index] = self.parent.slots[self.parent.select][i];
				self.parent.slots[self.parent.select][i] = x;
				self.parent.groups[self.index] = self.parent.groups[i];
				self.parent.groups[self.index].index = self.index;
				self.parent.groups[i] = self;
				self.index = i;
				self.parent:Changed();
				self.safeY = math.huge;
			end
		elseif args.Y > self:GetHeight() then
			local i = self.index + 1;

			if self.parent.groups[i] ~= nil then
				local x = self.parent.slots[self.parent.select][self.index];

				self.parent.slots[self.parent.select][self.index] = self.parent.slots[self.parent.select][i];
				self.parent.slots[self.parent.select][i] = x;
				self.parent.groups[self.index] = self.parent.groups[i];
				self.parent.groups[self.index].index = self.index;
				self.parent.groups[i] = self;
				self.index = i;
				self.parent:Changed();
				self.safeY = -math.huge;
			end
		end
	end
end

function PartyFrameOptionsGroup:MouseUp(args)
	self.dragging = false;
end

PartyFrameOptions = class(Turbine.UI.Control);

function PartyFrameOptions:Constructor(window)
	Turbine.UI.Control.Constructor(self);

	self.window = window;
	self.select = 0;
	self.effects = { };
	self.groups = { };

	self.effectsClipboard = nil;
	self.effectsIndex = { };
	self.effectsOptions = { };
	self.effectsMenu = Turbine.UI.ContextMenu();
	local items = self.effectsMenu:GetItems();
	self.effectsOptions.Cut = Turbine.UI.MenuItem("Cut");
	self.effectsOptions.Cut.parent = self;
	self.effectsOptions.Cut.Click = function(self, args)
		local x = self.parent.effectsIndex.x;
		local y = self.parent.effectsIndex.y;

		if y == nil then
			self.parent.effectsClipboard = self.parent.slots[self.parent.select][x];
			table.remove(self.parent.slots[self.parent.select], x);
		else
			self.parent.effectsClipboard = self.parent.slots[self.parent.select][x][y];
			table.remove(self.parent.slots[self.parent.select][x], y);
			if #self.parent.slots[self.parent.select][x] == 0 then
				table.remove(self.parent.slots[self.parent.select], x);
			end
		end
		self.parent:Changed();
	end;
	items:Add(self.effectsOptions.Cut);
	self.effectsOptions.Copy = Turbine.UI.MenuItem("Copy");
	self.effectsOptions.Copy.parent = self;
	self.effectsOptions.Copy.Click = function(self, args)
		local x = self.parent.effectsIndex.x;
		local y = self.parent.effectsIndex.y;
		local t;

		if y == nil then
			t = self.parent.slots[self.parent.select][x];
		else
			t = self.parent.slots[self.parent.select][x][y];
		end
		self.parent.effectsClipboard = { debuff = t.debuff, icon = t.icon, name = t.name };
	end;
	items:Add(self.effectsOptions.Copy);
	self.effectsOptions.Paste = Turbine.UI.MenuItem("Paste");
	self.effectsOptions.Paste.parent = self;
	self.effectsOptions.Paste.Click = function(self, args)
		local x = self.parent.effectsIndex.x;
		local y = self.parent.effectsIndex.y;
		local t = { };

		t.icon = self.parent.effectsClipboard.icon;
		t.debuff = self.parent.effectsClipboard.debuff;
		t.name = self.parent.effectsClipboard.name;

		if x == nil then
			if self.parent.slotsButton[self.parent.select].missing then
				table.insert(self.parent.slots[self.parent.select], { [1] = t });
			else
				table.insert(self.parent.slots[self.parent.select], t);
			end
		else
			if y == nil then
				table.insert(self.parent.slots[self.parent.select], x, t);
			else
				table.insert(self.parent.slots[self.parent.select][x], y, t);
			end
		end
		self.parent:Changed();
	end;
	items:Add(self.effectsOptions.Paste);
	self.effectsOptions.Delete = Turbine.UI.MenuItem("Delete");
	self.effectsOptions.Delete.parent = self;
	self.effectsOptions.Delete.Click = function(self, args)
		local x = self.parent.effectsIndex.x;
		local y = self.parent.effectsIndex.y;

		if y == nil then
			table.remove(self.parent.slots[self.parent.select], x);
		else
			table.remove(self.parent.slots[self.parent.select][x], y);
			if #self.parent.slots[self.parent.select][x] == 0 then
				table.remove(self.parent.slots[self.parent.select], x);
			end
		end
		self.parent:Changed();
	end;
	items:Add(self.effectsOptions.Delete);
	self.effectsOptions.Insert = Turbine.UI.MenuItem("Insert");
	self.effectsOptions.Insert.parent = self;
	self.effectsOptions.Insert.Click = function(self, args)
		local x = self.parent.effectsIndex.x;
		local y = self.parent.effectsIndex.y;

		if x == nil then
			if self.parent.slotsButton[self.parent.select].missing then
				table.insert(self.parent.slots[self.parent.select], { [1] = { debuff = false } });
			else
				table.insert(self.parent.slots[self.parent.select], { debuff = false });
			end
		else
			if y == nil then
				table.insert(self.parent.slots[self.parent.select], x, { debuff = false });
			else
				table.insert(self.parent.slots[self.parent.select][x], y, { debuff = false });
			end
		end
		self.parent:Changed();
	end;
	items:Add(self.effectsOptions.Insert);
	self.effectsOptions.Debuff = Turbine.UI.MenuItem("Debuff");
	self.effectsOptions.Debuff.parent = self;
	self.effectsOptions.Debuff.Click = function(self, args)
		local x = self.parent.effectsIndex.x;
		local y = self.parent.effectsIndex.y;

		if y == nil then
			self.parent.slots[self.parent.select][x].debuff = not self.parent.slots[self.parent.select][x].debuff;
		else
			self.parent.slots[self.parent.select][x][y].debuff = not self.parent.slots[self.parent.select][x][y].debuff;
		end
		self.parent:Changed();
	end;
	items:Add(self.effectsOptions.Debuff);
	self.effectsOptions.Save = Turbine.UI.MenuItem("Save");
	self.effectsOptions.Save.parent = self;
	self.effectsOptions.Save.Click = function(self, args)
		local x = self.parent.effectsIndex.x;
		local y = self.parent.effectsIndex.y;
		local t;

		if y == nil then
			t = self.parent.slots[self.parent.select][x];
		else
			t = self.parent.slots[self.parent.select][x][y];
		end
		
		for _, v in pairs(self.parent.logSaved) do
			if v.name == t.name and v.debuff == t.debuff and v.icon == t.icon then
				return;
			end
		end
		table.insert(self.parent.logSaved, { debuff = t.debuff, icon = t.icon, name = t.name });
		self.parent:SaveLogSaved();
	end;
	items:Add(self.effectsOptions.Save);
	self.effectsOptions.NewGroup = Turbine.UI.MenuItem("New Group");
	self.effectsOptions.NewGroup.parent = self;
	self.effectsOptions.NewGroup.Click = function(self, args)
		local x = self.parent.effectsIndex.x;
		local y = self.parent.effectsIndex.y;
		local n = #self.parent.slots[self.parent.select][x];

		table.insert(self.parent.slots[self.parent.select], x + 1, { });
		for i = y, n, 1 do
			table.insert(self.parent.slots[self.parent.select][x + 1], self.parent.slots[self.parent.select][x][i]);
			self.parent.slots[self.parent.select][x][i] = nil;
		end
		self.parent:Changed();
	end;
	items:Add(self.effectsOptions.NewGroup);

	self:SetBackColor(Turbine.UI.Color(1.0, 0.0, 0.0, 0.3));

	self.slots = self:LoadSlots();
	self.slotsButton = { };
	for i = -5, 5, 1 do
		if i ~= 0 then
			self.slotsButton[i] = PartyFrameOptionsButton(self, i, self.slots[i].missing);
		end
		if self.slots[i].effects == nil then
			self:Default(i);
		else
			self.slots[i] = self.slots[i].effects;
		end
	end
	self.slotsType = PartyFrameOptionsDropDown(self, 1, function(self, index) self.parent.slotsButton[self.parent.select]:SetMissing(index == 2); end, "Existent", "Missing");
	self.slotsType:SetSize(window.width, window.size - 2);
	self:UpdateSlots();

	self:LoadLogSaved();
	for k, v in pairs(self.logSaved) do
		self.slots[0][k] = v;
	end
	self.logButton = Turbine.UI.Lotro.Button();
	self.logButton.parent = self;
	self.logButton:SetParent(self);
	self.logButton:SetText("Log");
	self.logButton:SetSize(window.width, window.size - 2);
	self.logButton.MouseClick = function(self, args) self.parent:Select(0); end;
	self.logType = PartyFrameOptionsDropDown(self, 2, nil, "None", "Unknown", "All");
	self.logType:SetSize(window.width, window.size - 2);
	self.logBuffs = PartyFrameOptionsDropDown(self, 1, nil, "No Buffs", "All Buffs");
	self.logBuffs:SetSize(window.width, window.size - 2);
	self.logDebuffs = PartyFrameOptionsDropDown(self, 2, nil, "No Debuffs", "All Debuffs");
	self.logDebuffs:SetSize(window.width, window.size - 2);

	self.logClear = Turbine.UI.Lotro.Button();
	self.logClear.parent = self;
	self.logClear:SetParent(self);
	self.logClear:SetText("Clear");
	self.logClear:SetSize(window.width, window.size - 2);
	self.logClear.MouseClick = function(self, args)
		self.parent.slots[self.parent.select] = { };
		if self.parent.select == 0 then
			self.parent.logSaved = { };
		end
		self.parent:Changed();
	end;
	self.logDefault = Turbine.UI.Lotro.Button();
	self.logDefault.parent = self;
	self.logDefault:SetParent(self);
	self.logDefault:SetText("Default");
	self.logDefault:SetSize(window.width, window.size - 2);
	self.logDefault.MouseClick = function(self, args)
		if self.parent.select == 0 then
			self.parent.slots[0] = { };
			for k, v in pairs(self.parent.logSaved) do
				self.parent.slots[0][k] = v;
			end
		else
			self.parent:Default(self.parent.select);
		end
		self.parent:Changed();
	end;

	self.dragWindow = Turbine.UI.Window();
	self.dragWindow:SetSize(32, 32);
	self.dragWindow:SetBackColor(Turbine.UI.Color(1.0, 0.0, 0.0, 0.0));
	self.dragWindow:SetOpacity(0.7);

	window.options = self;
end

combatStates = { };
combatStates["Dazed"] = 0;
combatStates["Feared"] = 0;
combatStates["Knocked Down"] = 0;
combatStates["Knocked Out"] = 0;
combatStates["Recovering"] = 0;
combatStates["Rooted"] = 0;
combatStates["Stunned"] = 0;
combatStates["Vulnerable"] = 0;

debuffSlot = { };
debuffSlot[-4] = "<fear>";
debuffSlot[-3] = "<wound>";
debuffSlot[-2] = "<disease>";
debuffSlot[-1] = "<poison>";

buffSlot = { }
buffSlot[Turbine.Gameplay.Class.Burglar] = { };
buffSlot[Turbine.Gameplay.Class.Burglar][1] = { };
buffSlot[Turbine.Gameplay.Class.Burglar][1]["Signal: Conviction"] = 0;
buffSlot[Turbine.Gameplay.Class.Burglar][1]["Signal: Guile"] = 0;
buffSlot[Turbine.Gameplay.Class.Burglar][1]["Signal: Strength"] = 0;
buffSlot[Turbine.Gameplay.Class.Burglar][1]["Signal: Tactics"] = 0;

buffSlot[Turbine.Gameplay.Class.Captain] = { };
buffSlot[Turbine.Gameplay.Class.Captain][1] = { };
buffSlot[Turbine.Gameplay.Class.Captain][1]["Blade-brother"] = 0x411167F0;
buffSlot[Turbine.Gameplay.Class.Captain][1]["Shield-brother"] = 0x41104588;
buffSlot[Turbine.Gameplay.Class.Captain][1]["Song-brother"] = 0x411167EF;
buffSlot[Turbine.Gameplay.Class.Captain][1]["Watchful Shield-brother"] = 0x41104588;
buffSlot[Turbine.Gameplay.Class.Captain][2] = { };
buffSlot[Turbine.Gameplay.Class.Captain][2]["To Arms (Blade-brother)"] = 0x4107E9F2;
buffSlot[Turbine.Gameplay.Class.Captain][2]["To Arms (Shield-brother)"] = 0x411167F5;
buffSlot[Turbine.Gameplay.Class.Captain][2]["To Arms (Song-brother)"] = 0x411167F6;
buffSlot[Turbine.Gameplay.Class.Captain][2]["To Arms (Fellowship Blade-brother)"] = 0x4107E9F2;
buffSlot[Turbine.Gameplay.Class.Captain][2]["To Arms (Fellowship Shield-brother)"] = 0x411167F5;
buffSlot[Turbine.Gameplay.Class.Captain][2]["To Arms (Fellowship Song-brother)"] = 0x411167F6;
buffSlot[Turbine.Gameplay.Class.Captain][3] = { };
buffSlot[Turbine.Gameplay.Class.Captain][3]["War-cry"] = 0x4100560F;
buffSlot[Turbine.Gameplay.Class.Captain][4] = { };
buffSlot[Turbine.Gameplay.Class.Captain][4]["Make Haste"] = 0x4100371F;
buffSlot[Turbine.Gameplay.Class.Captain][4]["Rallying Cry"] = 0x41005609;
buffSlot[Turbine.Gameplay.Class.Captain][4]["Valiant Strike"] = 0x410E68E8;
buffSlot[Turbine.Gameplay.Class.Captain][4]["Words of Courage"] = 0x4100561B;

buffSlot[Turbine.Gameplay.Class.Champion] = { };

buffSlot[Turbine.Gameplay.Class.Guardian] = { };
buffSlot[Turbine.Gameplay.Class.Guardian][1] = { };
buffSlot[Turbine.Gameplay.Class.Guardian][1]["Protection"] = 0;
buffSlot[Turbine.Gameplay.Class.Guardian][1]["Protection by the Sword"] = 0x4111735D;
buffSlot[Turbine.Gameplay.Class.Guardian][1]["Shield Wall"] = 0;

buffSlot[Turbine.Gameplay.Class.Hunter] = { };
buffSlot[Turbine.Gameplay.Class.Hunter][1] = { };
buffSlot[Turbine.Gameplay.Class.Hunter][1]["Find the Path"] = 0x41008858;
buffSlot[Turbine.Gameplay.Class.Hunter][2] = { };
buffSlot[Turbine.Gameplay.Class.Hunter][2]["Cleanse Poison Resistance"] = 0x41008859;

buffSlot[Turbine.Gameplay.Class.LoreMaster] = { };
buffSlot[Turbine.Gameplay.Class.LoreMaster][1] = { };
buffSlot[Turbine.Gameplay.Class.LoreMaster][1]["Charged Air"] = 0x4110458C;
buffSlot[Turbine.Gameplay.Class.LoreMaster][2] = { };
buffSlot[Turbine.Gameplay.Class.LoreMaster][2]["Sign of Power: Righteousness"] = 0x4100881A;
buffSlot[Turbine.Gameplay.Class.LoreMaster][3] = { };
buffSlot[Turbine.Gameplay.Class.LoreMaster][3]["Sign of Power: Vigilance"] = 0x4100881B;

buffSlot[Turbine.Gameplay.Class.Minstrel] = { };
buffSlot[Turbine.Gameplay.Class.Minstrel][1] = { };
buffSlot[Turbine.Gameplay.Class.Minstrel][1]["Soliloquy of Spirit"] = 0x410E3A1F;
buffSlot[Turbine.Gameplay.Class.Minstrel][2] = { };
buffSlot[Turbine.Gameplay.Class.Minstrel][2]["Enlivening Chord"] = 0x41004E8F;
buffSlot[Turbine.Gameplay.Class.Minstrel][2]["Inspire Fellows Buff"] = 0x410087C7;
buffSlot[Turbine.Gameplay.Class.Minstrel][2]["The Melody of Battle"] = 0x410D73F5;
buffSlot[Turbine.Gameplay.Class.Minstrel][2]["Valar's Grace"] = 0x4100A4E2;
buffSlot[Turbine.Gameplay.Class.Minstrel][2]["Words of Courage: Fear Resistance"] = 0x410087D5;
buffSlot[Turbine.Gameplay.Class.Minstrel][3] = { };
buffSlot[Turbine.Gameplay.Class.Minstrel][3]["Fellowship's Heart"] = 0x4100A4E2;
buffSlot[Turbine.Gameplay.Class.Minstrel][4] = { };
buffSlot[Turbine.Gameplay.Class.Minstrel][4]["Tale of Battle"] = 0x410D715F;
buffSlot[Turbine.Gameplay.Class.Minstrel][4]["Tale of Frost and Flame"] = 0x410E3A26;
buffSlot[Turbine.Gameplay.Class.Minstrel][4]["Tale of Heroism"] = 0x410087D2;
buffSlot[Turbine.Gameplay.Class.Minstrel][4]["Tale of Warding"] = 0x410D7160;
buffSlot[Turbine.Gameplay.Class.Minstrel][4]["Tale of Warding and Hope"] = 0x410D7160;
buffSlot[Turbine.Gameplay.Class.Minstrel][4]["Improved Tale of Battle"] = 0x410D715F;
buffSlot[Turbine.Gameplay.Class.Minstrel][4]["Improved Tale of Warding"] = 0x410D7160;
buffSlot[Turbine.Gameplay.Class.Minstrel][5] = { };
buffSlot[Turbine.Gameplay.Class.Minstrel][5]["Anthem of Composure"] = 0x410087C6;
buffSlot[Turbine.Gameplay.Class.Minstrel][5]["Anthem of the Free People"] = 0x41005618;
buffSlot[Turbine.Gameplay.Class.Minstrel][5]["Anthem of Prowess"] = 0x410087C5;
buffSlot[Turbine.Gameplay.Class.Minstrel][5]["Anthem of War"] = 0x410056AE;

buffSlot[Turbine.Gameplay.Class.RuneKeeper] = { };
buffSlot[Turbine.Gameplay.Class.RuneKeeper][1] = { };
buffSlot[Turbine.Gameplay.Class.RuneKeeper][1]["Improved Rune-sign of Flame"] = 0x4110458E;
buffSlot[Turbine.Gameplay.Class.RuneKeeper][1]["Improved Rune-sign of Storm"] = 0x41104590;
buffSlot[Turbine.Gameplay.Class.RuneKeeper][1]["Improved Rune-sign of Winter"] = 0x4110458F;
buffSlot[Turbine.Gameplay.Class.RuneKeeper][1]["Prelude to Hope"] = 0x410E44D7;
buffSlot[Turbine.Gameplay.Class.RuneKeeper][1]["Rune-sign of Flame"] = 0x4110458E;
buffSlot[Turbine.Gameplay.Class.RuneKeeper][1]["Rune-sign of Storm"] = 0x41104590;
buffSlot[Turbine.Gameplay.Class.RuneKeeper][1]["Rune-sign of Winter"] = 0x4110458F;
buffSlot[Turbine.Gameplay.Class.RuneKeeper][2] = { };
buffSlot[Turbine.Gameplay.Class.RuneKeeper][2]["Writ of Health - Tier 1"] = 0x410EA25B;
buffSlot[Turbine.Gameplay.Class.RuneKeeper][2]["Writ of Health - Tier 2"] = 0x410EA25C;
buffSlot[Turbine.Gameplay.Class.RuneKeeper][2]["Writ of Health - Tier 3"] = 0x410EA25A;
buffSlot[Turbine.Gameplay.Class.RuneKeeper][3] = { };
buffSlot[Turbine.Gameplay.Class.RuneKeeper][3]["Mending Verse"] = 0x410E44D4;
buffSlot[Turbine.Gameplay.Class.RuneKeeper][4] = { };
buffSlot[Turbine.Gameplay.Class.RuneKeeper][4]["Essay of Exaltation"] = 0x410F3580;
buffSlot[Turbine.Gameplay.Class.RuneKeeper][4]["Glorious Foreshadowing"] = 0x410F23E9;
buffSlot[Turbine.Gameplay.Class.RuneKeeper][4]["Rousing Words"] = 0x410E44CE;
buffSlot[Turbine.Gameplay.Class.RuneKeeper][4]["That Which Does Not Kill Us"] = 0x410F23E6;
buffSlot[Turbine.Gameplay.Class.RuneKeeper][4]["Wondrous Foreshadowing"] = 0x410F23EA;
buffSlot[Turbine.Gameplay.Class.RuneKeeper][5] = { };
buffSlot[Turbine.Gameplay.Class.RuneKeeper][5]["Do not fall to Fire"] = 0x410EA22F;
buffSlot[Turbine.Gameplay.Class.RuneKeeper][5]["Do not fall to Frost"] = 0x410EA234;
buffSlot[Turbine.Gameplay.Class.RuneKeeper][5]["Do not fall to Lightning"] = 0x410EA235;
buffSlot[Turbine.Gameplay.Class.RuneKeeper][5]["Do Not Fall This Day"] = 0;
buffSlot[Turbine.Gameplay.Class.RuneKeeper][5]["Shall Not Fall This Day"] = 0x410E44DC;
buffSlot[Turbine.Gameplay.Class.RuneKeeper][5]["Their Weapons Shall Not Harm Us"] = 0;

buffSlot[Turbine.Gameplay.Class.Warden] = { };
buffSlot[Turbine.Gameplay.Class.Warden][1] = { };
buffSlot[Turbine.Gameplay.Class.Warden][1]["Conviction"] = 0x410EA2A7;
buffSlot[Turbine.Gameplay.Class.Warden][2] = { };
buffSlot[Turbine.Gameplay.Class.Warden][2]["Conviction - Physical Mastery"] = 0x410E8292;
buffSlot[Turbine.Gameplay.Class.Warden][2]["Conviction - Threat Reduction"] = 0x410E87A3;

captainSlot = { };
captainSlot[1] = { };
captainSlot[1][1] = { name = "Focus", icon = 0x410087BB };
captainSlot[1][2] = { name = "On Guard", icon = 0x410087B5 };
captainSlot[1][3] = { name = "Relentless Attack", icon = 0x410087B7 };
captainSlot[2] = { };
captainSlot[2][1] = { name = "Motivated", icon = 0x410087B4 };
captainSlot[3] = { };
captainSlot[3][1] = { name = "Strength of Will (Shield-brother)", icon = 0x410D3FFA };
captainSlot[3][2] = { name = "Strength of Will (Blade-brother)", icon = 0x411167F4 };
captainSlot[3][3] = { name = "Strength of Will (Song-brother)", icon = 0x411167F3 };
captainSlot[3][4] = { name = "Strength of Will (Fellowship Blade-brother)", icon = 0x411167F4 };
captainSlot[3][5] = { name = "Strength of Will (Fellowship Shield-brother)", icon = 0x410D3FFA };
captainSlot[3][6] = { name = "Strength of Will (Fellowship Song-brother)", icon = 0x411167F3 };
captainSlot[4] = { };
captainSlot[4][1] = { name = "Inspired (Blade-brother)", icon = 0x411167F2 };
captainSlot[4][2] = { name = "Inspired (Shield-brother)", icon = 0x4107E9F1 };
captainSlot[4][3] = { name = "Inspired (Song-brother)", icon = 0x411167F1 };
captainSlot[4][4] = { name = "Inspired (Fellowship Blade-brother)", icon = 0x411167F2 };
captainSlot[4][5] = { name = "Inspired (Fellowship Shield-brother)", icon = 0x4107E9F1 };
captainSlot[4][6] = { name = "Inspired (Fellowship Song-brother)", icon = 0x411167F1 };
captainSlot[5] = { };
captainSlot[5][1] = { name = "Light of Elendil", icon = 0x4109B53E };

function PartyFrameOptions:Default(i)
	self.slots[i] = { };
	if i ~= 0 then
		self.slotsButton[i].missing = false;
	end

	if i == -5 then
		for k, v in pairs(combatStates) do
			local d = (k ~= "Recovering");

			if v ~= 0 then
				table.insert(self.slots[i], { icon = v, debuff = d, name = k });
			else
				table.insert(self.slots[i], { debuff = d, name = k });
			end
		end
	elseif i < 0 then
		table.insert(self.slots[i], { debuff = true, name = debuffSlot[i] });
	elseif i == 5 and self.window.class == Turbine.Gameplay.Class.Captain then
		self.slotsButton[i].missing = true;
		for k, v in ipairs(captainSlot) do
			self.slots[i][k] = { };
			for _, v1 in ipairs(v) do
				table.insert(self.slots[i][k], { icon = v1.icon, debuff = false, name = v1.name });
			end
		end
	elseif buffSlot[self.window.class][i] ~= nil then
		for k, v in pairs(buffSlot[self.window.class][i]) do
			if v ~= 0 then
				table.insert(self.slots[i], { icon = v, debuff = false, name = k });
			else
				table.insert(self.slots[i], { debuff = false, name = k });
			end
		end
	end

	if i ~= 0 then
		self.slotsButton[i]:Colorize();
	end
	if i == 0 or not self.slotsButton[i].missing then
		table.sort(self.slots[i], function(a, b) return (not a.name and "" or a.name) < (not b.name and "" or b.name); end);
	end
end

function PartyFrameOptions:Changed()
	self:SizeChanged();
	self:SaveLogSaved();
	self:SaveSlots();
	self:UpdateSlots();
end

function PartyFrameOptions:UpdateSlots()
	local name;

	effectGroup = { };
	effectSlot = { };

	for i = -5, 5, 1 do
		if i ~= 0 then
			if self.slotsButton[i].missing then
				effectGroup[i] = { };
				for k, v in ipairs(self.slots[i]) do
					effectGroup[i][k] = { };
					effectGroup[i][k].icon = v[1].icon;
					effectGroup[i][k].effects = { };

					for _, v1 in pairs(v) do
						if v1.debuff then
							name = "D";
						else
							name = "B";
						end

						if v1.icon ~= nil then
							name = name .. "/" .. v1.icon;
						end
						if v1.name ~= nil then
							name = name .. "/" .. v1.name;
						end

						if string.len(name) > 1 then
							effectSlot[name] = i;
							effectGroup[i][k].effects[name] = true;
						end
					end
				end
			else
				for _, v in pairs(self.slots[i]) do
					if v.debuff then
						name = "D";
					else
						name = "B";
					end

					if v.icon ~= nil then
						name = name .. "/" .. v.icon;
					end
					if v.name ~= nil then
						name = name .. "/" .. v.name;
					end

					if string.len(name) > 1 then
						effectSlot[name] = i;
					end
				end
			end
		end
	end

	for k, v in pairs(self.window.actors) do
		v:SetupEffects();
	end
end

function PartyFrameOptions:SaveSlots()
	local t = { };

	for k, v in pairs(self.slots) do
		if k ~= 0 then
			t[k] = { };
			t[k].missing = self.slotsButton[k].missing;
			t[k].effects = v;
		end
	end
	Turbine.PluginData.Save(Turbine.DataScope.Character, "PartyFrame_Slots", t);
end

function PartyFrameOptions:LoadSlots()
	local t = Turbine.PluginData.Load(Turbine.DataScope.Character, "PartyFrame_Slots");

	if t == nil then
		t = { };
	end

	for i = -5, 5, 1 do
		if t[i] == nil then
			t[i] = { };
		end
	end

	return t;
end

function PartyFrameOptions:SaveLogSaved()
	Turbine.PluginData.Save(Turbine.DataScope.Account, "PartyFrame_LogSaved", self.logSaved);
end

function PartyFrameOptions:LoadLogSaved()
	self.logSaved = Turbine.PluginData.Load(Turbine.DataScope.Account, "PartyFrame_LogSaved");
	if self.logSaved == nil then
		self.logSaved = { };
	end
end

function PartyFrameOptions:GetIndices(i)
	local x = nil;
	local y = nil;

	if self.select == 0 then
		x = i;
	elseif self.slotsButton[self.select].missing then
		for k, v in ipairs(self.slots[self.select]) do
			if i <= #v then
				x = k;
				y = i;
				break;
			end
			i = i - #v;
		end
	else
		if self.slots[self.select][i] ~= nil then
			x = i;
		end
	end

	return x, y;
end

function PartyFrameOptions:Select(i)
	self.select = i;

	for _, v in pairs(self.slotsButton) do
		v:Colorize();
	end

	if self.select ~= 0 then
		if self.slotsButton[self.select].missing then
			self.slotsType:SetIndex(2);
		else
			self.slotsType:SetIndex(1);
		end
	end

	self:SizeChanged();
end

function PartyFrameOptions:SizeChanged(args)
	local w, h = self:GetSize();
	local x;
	local y = 0;

	y = y + 10;
	x = w - (self.window.size * 10 + self.window.width);
	if x > 0 then
		x = x / 2;
		for i = -5, 5, 1 do
			if i == 0 then
				self.logButton:SetPosition(x, y);
				x = x + self.window.width;
			else
				self.slotsButton[i]:SetPosition(x, y);
				x = x + self.window.size;
			end
		end
		y = y + self.window.size;
	else
		x = (w - (self.window.size * 5)) / 2;
		for i = -5, -1, 1 do
			self.slotsButton[i]:SetPosition(x, y);
			x = x + self.window.size;
		end
		y = y + self.window.size;

		x = (w - self.window.width) / 2;
		self.logButton:SetPosition(x, y);
		y = y + self.window.size;

		x = (w - (self.window.size * 5)) / 2;
		for i = 1, 5, 1 do
			self.slotsButton[i]:SetPosition(x, y);
			x = x + self.window.size;
		end
		y = y + self.window.size;
	end
	y = y + 10;

	if self.select == 0 then
		self.logType:SetVisible(true);
		self.logBuffs:SetVisible(true);
		self.logDebuffs:SetVisible(true);
		self.slotsType:SetVisible(false);
		x = w - (self.window.width * 3 + 10);
		if x > 0 then
			x = x / 2;
			self.logType:SetPosition(x, y);
			x = x + self.window.width + 5
			self.logBuffs:SetPosition(x, y);
			x = x + self.window.width + 5
			self.logDebuffs:SetPosition(x, y);
			y = y + self.window.size;
		else
			x = (w - self.window.width) / 2;
			self.logType:SetPosition(x, y);
			y = y + self.window.size;
			self.logBuffs:SetPosition(x, y);
			y = y + self.window.size;
			self.logDebuffs:SetPosition(x, y);
			y = y + self.window.size;
		end
	else
		self.logType:SetVisible(false);
		self.logBuffs:SetVisible(false);
		self.logDebuffs:SetVisible(false);
		self.slotsType:SetVisible(true);
		x = (w - self.window.width) / 2;
		self.slotsType:SetPosition(x, y);
		y = y + self.window.size;
	end
	y = y + 10;

	for k, v in ipairs(self.effects) do
		v:SetVisible(false);
	end
	for k, v in ipairs(self.groups) do
		v:SetVisible(false);
	end
	if self.select == 0 then
		for k, v in ipairs(self.slots[0]) do
			if self.effects[k] == nil then
				self.effects[k] = PartyFrameOptionsEffect(self, k, v.icon, v.debuff, v.name, true);
			else
				self.effects[k]:SetIcon(v.icon);
				self.effects[k]:SetDebuff(v.debuff);
				self.effects[k]:SetName(v.name, true);
			end
			self.effects[k]:SetSize(w - (6 + self.window.size), self.window.size + 2);
			self.effects[k]:SetPosition(4 + self.window.size, y);
			self.effects[k]:SetVisible(true);
			y = y + self.window.size + 2;
		end
	else
		local i = 1;

		if self.slotsButton[self.select].missing then
			for k, v in ipairs(self.slots[self.select]) do
				if self.groups[k] == nil then
					self.groups[k] = PartyFrameOptionsGroup(self, k);
				end
				self.groups[k]:Setup(#v);
				self.groups[k]:SetPosition(2, y + 2);

				for _, v1 in ipairs(v) do
					if self.effects[i] == nil then
						self.effects[i] = PartyFrameOptionsEffect(self, i, v1.icon, v1.debuff, v1.name, false);
					else
						self.effects[i]:SetIcon(v1.icon);
						self.effects[i]:SetDebuff(v1.debuff);
						self.effects[i]:SetName(v1.name, false);
					end
					self.effects[i]:SetSize(w - (6 + self.window.size), self.window.size + 2);
					self.effects[i]:SetPosition(4 + self.window.size, y);
					self.effects[i]:SetVisible(true);
					i = i + 1;
					y = y + self.window.size + 2;
				end
			end
		else
			for _, v in ipairs(self.slots[self.select]) do
				if self.effects[i] == nil then
					self.effects[i] = PartyFrameOptionsEffect(self, i, v.icon, v.debuff, v.name, false);
				else
					self.effects[i]:SetIcon(v.icon);
					self.effects[i]:SetDebuff(v.debuff);
					self.effects[i]:SetName(v.name, false);
				end
				self.effects[i]:SetSize(w - (6 + self.window.size), self.window.size + 2);
				self.effects[i]:SetPosition(4 + self.window.size, y);
				self.effects[i]:SetVisible(true);
				i = i + 1;
				y = y + self.window.size + 2;
			end
		end

		if self.effects[i] == nil then
			self.effects[i] = PartyFrameOptionsEffect(self, i, nil, false, nil, false);
		else
			self.effects[i]:SetIcon(nil);
			self.effects[i]:SetDebuff(false);
			self.effects[i]:SetName(nil, false);
		end
		self.effects[i]:SetSize(w - (6 + self.window.size), self.window.size + 2);
		self.effects[i]:SetPosition(4 + self.window.size, y);
		self.effects[i]:SetVisible(true);
		i = i + 1;
		y = y + self.window.size + 2;
	end
	y = y + 10;

	x = w - (self.window.width * 2 + 5);
	if x > 0 then
		x = x / 2;
		self.logClear:SetPosition(x, y);
		x = x + self.window.width + 5
		self.logDefault:SetPosition(x, y);
		y = y + self.window.size;
	else
		x = (w - self.window.width) / 2;
		self.logClear:SetPosition(x, y);
		y = y + self.window.size;
		self.logDefault:SetPosition(x, y);
		y = y + self.window.size;
	end
	y = y + 10;

	self:SetHeight(y);
end

PartyFrameCommand = class(Turbine.ShellCommand);

function PartyFrameCommand:Constructor(window)
	Turbine.ShellCommand.Constructor(self);

	self.window = window;
end

function PartyFrameCommand:Execute(command, arguments)
	local party = self.window.player:GetParty();

	if party ~= self.window.party then
		self.window:UpdateParty();
	end

	if arguments == "clear" then
		for k, v in pairs(self.window.players) do
			v.rolls = { };
			v.rollOrder = nil;
			v:UpdateName();
		end
	elseif self.window.party ~= nil then
		self.window:PartyMemberChanged(arguments);
	end
end

function PartyFrameCommand:GetShortHelp()
	return "PartyFrame";
end

plugin.window = PartyFrameWindow(22, 100);
plugin.options = PartyFrameOptions(plugin.window);
plugin.command = PartyFrameCommand(plugin.window);
Turbine.Shell.AddCommand("partyframe", plugin.command);
Turbine.Shell.AddCommand("pf", plugin.command);

function plugin:Load(args)
	local state;

	state = Turbine.PluginData.Load(Turbine.DataScope.Account, "PartyFrame");
	if (Turbine.Engine.GetLocalTime() - state.time) < 15 then
		self.window:PartyMemberChanged(state.order);
	end
end

function plugin:Unload(args)
	local state = { };

	state.time = Turbine.Engine.GetLocalTime();
	state.order = "";
	for i = 1, #self.window.fellowship, 1 do
		if i > 1 then
			state.order = state.order .. " ";
		end

		for j = 1, #self.window.fellowship[i], 1 do
			if j > 1 then
				state.order = state.order .. ",";
			end
			state.order = state.order .. self.window.fellowship[i][j].actor:GetName();
		end
	end
	Turbine.PluginData.Save(Turbine.DataScope.Account, "PartyFrame", state);
end

function plugin:GetOptionsPanel()
	return self.options;
end